﻿local raceFrame;
local achieveFrame;
local travelFrame;
local linksFrame;
local itemsFrame;
local optionsFrame;
local _, _, _, patchVersion = GetBuildInfo()

local function makeLabel(frame, size, text)
	local label = frame:CreateFontString()
	label:SetFont("Fonts\\FRIZQT__.TTF", size)
	label:SetText(text)
	return label
end

local function setTexture(frame, c1, c2, c3, a)
	local t=frame:CreateTexture()
	t:SetTexture(c1, c2, c3, a or 1)
	t:SetAllPoints(frame)
	frame.texture = t
end

local function changeTexture(frame, c1, c2, c3, a)
	local t=frame.texture
	t:SetTexture(c1, c2, c3, a or 1)
	frame.texture = t
end

local version = GetAddOnMetadata("Excavatinator", "Version")
if tonumber(version) < 1.5 then
	message("This update to Excavatinator requires a client restart.")
end

-- Build artifact list.
local artifactIDs = {
	{ -- 1 (Dwarf)
		common = {
			["Belt Buckle with Anvilmar Crest"] = 63113,
			["Bodacious Door Knocker"] = 64339,
			["Bone Gaming Dice"] = 63112,
			["Boot Heel with Scrollwork"] = 64340,
			["Ceramic Funeral Urn"] = 63409,
			["Dented Shield of Horuz Killcrow"] = 64362,
			["Dwarven Baby Socks"] = 66054,
			["Golden Chamber Pot"] = 64342,
			["Ironstar's Petrified Shield"] = 64344,
			["Mithril Chain of Angerforge"] = 64368,
			["Moltenfist's Jeweled Goblet"] = 63414,
			["Notched Sword of Tunadil the Redeemer"] = 64337,
			["Pewter Drinking Cup"] = 63408,
			["Pipe of Franclorn Forgewright"] = 64659,
			["Scepter of Bronzebeard"] = 64487,
			["Scepter of Charlga Razorflank"] = 64367,
			["Scorched Staff of Shadow Priest Anund"] = 64366,
			["Silver Kris of Korl"] = 64483,
			["Silver Neck Torc"] = 63411,
			["Skull Staff of Shadowforge"] = 64371,
			["Spiked Gauntlets of Anvilrage"] = 64485,
			["Stone Gryphon"] = 63410,
			["Warmaul of Burningeye"] = 64484,
			["Winged Helm of Corehammer"] = 64343,
			["Wooden Whistle"] = 63111,
			["Word of Empress Zoe"] = 64486,
			["Worn Hunting Knife"] = 63110,
		},
		rare = {
			["Chalice of the Mountain Kings"] = 64373,
			["Clockwork Gnome"] = 64372,
			["Staff of Sorcerer-Thane Thaurissan"] = 64489,
			["The Innkeeper's Daughter"] = 64488,
		},
		special = {
			["n"] = 0,
		},
	},
	{ -- 2 (Draenei)
		common = {
			["Anklet with Golden Bells"] = 64440,
			["Baroque Sword Scabbard"] = 64453,
			["Carved Harp of Exotic Wood"] = 64442,
			["Dignified Portrait"] = 64455,
			["Fine Crystal Candelabra"] = 64454,
			["Plated Elekk Goad"] = 64458,
			["Scepter of the Nathrezim"] = 64444,
			["Strange Silver Paperweight"] = 64443,
		},
		rare = {
			["Arrival of the Naaru"] = 64456,
			["The Last Relic of Argus"] = 64457,
		},
		special = {
			["n"] = 0,
		},
	},
	{ -- 3 (Fossil)
		common = {
			["Ancient Shark Jaws"] = 64355,
			["Beautiful Preserved Fern"] = 63121,
			["Black Trilobite"] = 63109,
			["Devilsaur Tooth"] = 64349,
			["Feathered Raptor Arm"] = 64385,
			["Imprint of a Kraken Tentacle"] = 64473,
			["Insect in Amber"] = 64350,
			["Proto-Drake Skeleton"] = 64468,
			["Shard of Petrified Wood"] = 66056,
			["Strange Velvet Worm"] = 66057,
			["Twisted Ammonite Shell"] = 63527,
			["Vicious Ancient Fish"] = 64387,
		},
		rare = {
			["Fossilized Hatchling"] = 60955,
			["Fossilized Raptor"] = 60954,
		},
		special = {
			["n"] = 0,
		},
	},
	{ -- 4 (Night Elf)
		common = {
			["Carcanet of the Hundred Magi"] = 64647,
			["Chest of Tiny Glass Animals"] = 64379,
			["Cloak Clasp with Antlers"] = 63407,
			["Coin from Eldre'Thalas"] = 63525,
			["Cracked Crystal Vial"] = 64381,
			["Delicate Music Box"] = 64357,
			["Green Dragon Ring"] = 63528,
			["Hairpin of Silver and Malachite"] = 64356,
			["Highborne Pyxis"] = 63129,
			["Inlaid Ivory Comb"] = 63130,
			["Kaldorei Amphora"] = 64354,
			["Necklace with Elune Pendant"] = 66055,
			["Scandalous Silk Nightgown"] = 63131,
			["Scepter of Xavius"] = 64382,
			["Shattered Glaive"] = 63526,
			["Silver Scroll Case"] = 64648,
			["String of Small Pink Pearls"] = 64378,
			["Umbra Crescent"] = 64650,
		},
		rare = {
			["Bones of Transformation"] = 64646,
			["Druid and Priest Statue Set"] = 64361,
			["Highborne Soul Mirror"] = 64358,
			["Kaldorei Wind Chimes"] = 64383,
			["Queen Azshara's Dressing Gown"] = 64643,
			["Tyrande's Favorite Doll"] = 64645,
			["Wisp Amulet"] = 64651,
		},
		special = {
			["n"] = 0,
		},
	},
	{ -- 5 (Nerubian)
		common = {
			["Ewer of Jormungar Blood"] = 64479,
			["Gruesome Heart Box"] = 64477,
			["Infested Ruby Ring"] = 64476,
			["Scepter of Nezar'Azret"] = 64475,
			["Six-Clawed Cornice"] = 64478,
			["Spidery Sundial"] = 64474,
			["Vizier's Scrawled Streamer"] = 64480,
		},
		rare = {
			["Blessing of the Old God"] = 64481,
			["Puzzle Box of Yogg-Saron"] = 64482,
		},
		special = {
			["n"] = 0,
		},
	},
	{ -- 6 (Orc)
		common = {
			["Fiendish Whip"] = 64436,
			["Fierce Wolf Figurine"] = 64421,
			["Gray Candle Stub"] = 64418,
			["Maul of Stone Guard Mur'og"] = 64417,
			["Rusted Steak Knife"] = 64419,
			["Scepter of Nekros Skullcrusher"] = 64420,
			["Skull Drinking Cup"] = 64438,
			["Tile of Glazed Clay"] = 64437,
			["Tiny Bronze Scorpion"] = 64389,
		},
		rare = {
			["Headdress of the First Shaman"] = 64644,
		},
		special = {
			["n"] = 0,
		},
	},
	{ -- 7 (Tol["vir)
		common = {
			["Canopic Jar"] = 64657,
			["Castle of Sand"] = 64652,
			["Cat Statue with Emerald Eyes"] = 64653,
			["Engraved Scimitar Hilt"] = 64656,
			["Sketch of a Desert Palace"] = 64658,
			["Soapstone Scarab Necklace"] = 64654,
			["Tiny Oasis Mosaic"] = 64655,
		},
		rare = {
			["Crawling Claw"] = 60847,
			["Pendant of the Scarab Storm"] = 64881,
			["Ring of the Boy Emperor"] = 64904,
			["Scepter of Azj'Aqir"] = 64883,
			["Scimitar of the Sirocco"] = 64885,
			["Staff of Ammunae"] = 64880,
		},
		special = {
			["n"] = 1,
			["Recipe: Vial of the Sands"] = 67538,
		},
	},
	{ -- 8 (Troll)
		common = {
			["Atal'ai Scepter"] = 64348,
			["Bracelet of Jade and Coins"] = 64346,
			["Cinnabar Bijou"] = 63524,
			["Drakkari Sacrificial Knife"] = 64375,
			["Eerie Smolderthorn Idol"] = 63523,
			["Feathered Gold Earring"] = 63413,
			["Fetish of Hir'eek"] = 63120,
			["Fine Bloodscalp Dinnerware"] = 66058,
			["Gahz'rilla Figurine"] = 64347,
			["Jade Asp with Ruby Eyes"] = 63412,
			["Lizard Foot Charm"] = 63118,
			["Skull-Shaped Planter"] = 64345,
			["Tooth with Gold Filling"] = 64374,
			["Zandalari Voodoo Doll"] = 63115,
		},
		rare = {
			["Zin'rokh, Destroyer of Worlds"] = 64377,
		},
		special = {
			["n"] = 0,
		},
	},
	{ -- 9 (Vrykul)
		common = {
			["Fanged Cloak Pin"] = 64464,
			["Flint Striker"] = 64462,
			["Intricate Treasure Chest Key"] = 64459,
			["Scramseax"] = 64461,
			["Thorned Necklace"] = 64467,
		},
		rare = {
			["Nifflevar Bearded Axe"] = 64460,
		},
		special = {
			["n"] = 0,
		},
	},
	races = 9,
}

local artifactsPatchedIDs = {
	[40100] = {
		{ -- 1 (Dwarf)
			common = {
			},
			rare = {
			},
			special = { ["n"] = 0, },
		},
		{ -- 2 (Draenei)
			common = {
			},
			rare = {
			},
			special = { ["n"] = 0, },
		},
		{ -- 3 (Fossil)
			common = {
			},
			rare = {
				["Pterrordax Hatchling"] = 69821,
				["Ancient Amber"] = 69776,
				["Extinct Turtle Shell"] = 69764,
			},
			special = { ["n"] = 0, },
		},
		{ -- 4 (Night Elf)
			common = {
			},
			rare = {
			},
			special = { ["n"] = 0, },
		},
		{ -- 5 (Nerubian)
			common = {
			},
			rare = {
			},
			special = { ["n"] = 0, },
		},
		{ -- 6 (Orc)
			common = {
			},
			rare = {
			},
			special = { ["n"] = 0, },
		},
		{ -- 7 (Tol'vir)
			common = {
			},
			rare = {
			},
			special = { ["n"] = 0, },
		},
		{ -- 8 (Troll)
			common = {
			},
			rare = {
				["Voodoo Figurine"] = 69824,
				["Haunted War Drum"] = 69777,
			},
			special = { ["n"] = 0, },
		},
		{ -- 9 (Vrykul)
			common = {
			},
			rare = {
				["Vrykul Drinking Horn"] = 69775,
			},
			special = { ["n"] = 0, },
		},
		patch = "4.1.0",
		races = 9,
	},
}

local linkedProjects = {
	-- ["Project name"] = "Item name"
	-- English
	["Mummified Monkey Paw"] = "Crawling Claw",
	-- French
	["Dé en os"] = "Dé de jeu en os",
	["Couteau de chasseur usé"] = "Couteau de chasse usé",
	["Étrange presse-papier en argent"] = "Etrange presse-papier en argent",
	["Épingle à cheveux en argent et malachite"] = "Epingle à cheveux en argent et malachite",
}
local linkedProjectsByItemName = {}

local potentialLinks = {}

local continents = {
	{ 1, 3, 4, 7, 8 }, -- Kalimdor
	{ 1, 3, 4, 8 }, -- Eastern Kingdoms
	{ 2, 6 }, -- Outland
	{ 4, 5, 8, 9 }, -- Northrend
}
continents[0] = { 1, 2, 3, 4, 5, 6, 7, 8, 9 } -- Failsafe

local travel = {
	spells = {
		-- spell ID, cooldown, reagent (0 if none) (in seconds)
		{ 556, 900, 0 }, -- Astral Recall
		{ 3561, 0, 17031 }, -- Teleport: Stormwind
		{ 3562, 0, 17031 }, -- Teleport: Ironforge
		{ 3565, 0, 17031 }, -- Teleport: Darnassus
		{ 32271, 0, 17031 }, -- Teleport: Exodar
		{ 49359, 0, 17031 }, -- Teleport: Theramore
		{ 3567, 0, 17031 }, -- Teleport: Orgrimmar
		{ 3563, 0, 17031 }, -- Teleport: Undercity
		{ 3566, 0, 17031 }, -- Teleport: Thunder Bluff
		{ 32272, 0, 17031 }, -- Teleport: Silvermoon
		{ 49358, 0, 17031 }, -- Teleport: Stonard
		{ 33690, 0, 17031 }, -- Teleport: Shattrath (Alliance)
		{ 35715, 0, 17031 }, -- Teleport: Shattrath (Horde)
		{ 53140, 0, 17031 }, -- Teleport: Dalaran
		{ 88342, 0, 17031 }, -- Teleport: Tol Barad (Alliance)
		{ 88344, 0, 17031 }, -- Teleport: Tol Barad (Horde)
		{ 18960, 0, 0 }, -- Teleport: Moonglade
		{ 50977, 60, 0 }, -- Death Gate
	},
	items = {
		-- item ID, spell ID, cooldown (in seconds), consumed (boolean)
		{ 6948, 8690, 1800, false }, -- Hearthstone
		{ 28585, 39937, 1800, false }, -- Ruby Slippers
		{ 54452, 75136, 1800, false }, -- Ethereal Portal
		{ 64488, 94719, 1800, false }, -- The Innkeeper's Daughter
		{ 37118, 48129, 1200, true }, -- Scroll of Recall
		{ 44314, 60320, 1200, true }, -- Scroll of Recall II
		{ 44315, 60321, 1200, true }, -- Scroll of Recall III
		{ 18986, 23453, 14400, false }, -- Ultrasafe Transporter - Gadgetzan
		{ 30544, 36941, 14400, false }, -- Ultrasafe Transporter - Toshley's Station
		{ 18984, 23442, 14400, false }, -- Dimensional Ripper - Everlook
		{ 30542, 36890, 14400, false }, -- Dimensional Ripper - Area 52
		{ 48933, 67833, 14400, false }, -- Wormhole Generator: Northrend
		{ 50287, 71436, 86400, false }, -- Boots of the Bay
		{ 63352, 89157, 28800, false }, -- Shroud of Cooperation (Stormwind)
		{ 63353, 89158, 28800, false }, -- Shroud of Cooperation (Orgrimmar)
		{ 63206, 89157, 14400, false }, -- Wrap of Unity (Stormwind)
		{ 63207, 89158, 14400, false }, -- Wrap of Unity (Orgrimmar)
		{ 65360, 89157, 7200, false }, -- Cloak of Coordination (Stormwind)
		{ 65274, 89158, 7200, false }, -- Cloak of Coordination (Orgrimmar)
		{ 52251, 73324, 3600, false }, -- Jaina's Locket
		{ 40586, 54406, 1800, false }, -- Band of the Kirin Tor
		{ 45688, 54406, 1800, false }, -- Inscribed Band of the Kirin Tor
		{ 48954, 54406, 1800, false }, -- Etched Band of the Kirin Tor
		{ 51560, 54406, 1800, false }, -- Runed Band of the Kirin Tor
		{ 44935, 54406, 1800, false }, -- Ring of the Kirin Tor
		{ 45690, 54406, 1800, false }, -- Inscribed Ring of the Kirin Tor
		{ 48956, 54406, 1800, false }, -- Etched Ring of the Kirin Tor
		{ 51559, 54406, 1800, false }, -- Runed Ring of the Kirin Tor
		{ 44934, 54406, 1800, false }, -- Loop of the Kirin Tor
		{ 45689, 54406, 1800, false }, -- Inscribed Loop of the Kirin Tor
		{ 48955, 54406, 1800, false }, -- Etched Loop of the Kirin Tor
		{ 51558, 54406, 1800, false }, -- Runed Loop of the Kirin Tor
		{ 40585, 54406, 1800, false }, -- Signet of the Kirin Tor
		{ 45691, 54406, 1800, false }, -- Inscribed Signed of the Kirin Tor
		{ 48957, 54406, 1800, false }, -- Etched Signet of the Kirin Tor
		{ 51557, 54406, 1800, false }, -- Runed Signet of the Kirin Tor
	},
}

onUseItems = {
	-- item ID, spell ID, cooldown (in seconds), consumed (boolean)
	-- Dwarf
	{ 64373, 94690, 600, false }, -- Chalice of the Mountain Kings
	{ 64488, 94719, 1800, false }, -- The Innkeeper's Daughter
	-- Draenei
	{ 64457, 82674, 43200, false }, -- The Last Relic of Argus
	{ 64456, 91752, 180, false }, -- Arrival of the Naaru
	-- Fossil
	{ 69776, 98445, 1800, false }, -- Ancient Amber
	-- Night Elf
	{ 64358, 95007, 600, false }, -- Highborne Soul Mirror
	{ 64383, 91749, 60, false }, -- Kaldorei Wind Chimes
	{ 64361, 91743, 900, false }, -- Druid and Priest Statue Set
	{ 64646, 91760, 600, false }, -- Bones of Transformation
	{ 64651, 91771, 600, false }, -- Wisp Amulet
	-- Nerubian
	{ 64481, 91754, 600, false }, -- Blessing of the Old God
	{ 64482, 91756, 60, false }, -- Puzzle Box of Yogg-Saron
	-- Orc
	-- Nothing for orc...
	-- Tol'vir
	{ 64881, 92146, 600, false }, -- Pendant of the Scarab Storm
	-- Troll
	{ 69777, 98447, 1800, false }, -- Haunted War Drum
	-- Vrykul
	{ 69775, 98444, 0, false }, -- Vrykul Drinking Horn
	-- Learnable items (pets, mounts)
	{ 64372, 90523, 0, true }, -- Clockwork Gnome
	{ 69821, 98571, 0, true }, -- Pterrordax Hatchling
	{ 60955, 84752, 0, true }, -- Fossilized Hatchling
	{ 60954, 84751, 0, true }, -- Fossilized Raptor
	{ 60847, 84263, 0, true }, -- Crawling Claw
	{ 64883, 92155, 0, true }, -- Scepter of Azj'Aqir
	{ 69824, 98587, 0, true }, -- Voodoo Figurine
}
-- Make sure to cache the necessary information
for i=1, #travel.items do
	GetItemIcon(travel.items[i][1])
	GetItemInfo(travel.items[i][1])
end
for i=1, #travel.spells do
	GetSpellInfo(travel.spells[i][1])
end

local artifacts, artifactsPatched = {}, {}
local getProgressCallback, excavatinator;

local historyFrame = CreateFrame("Frame")

local function deleteArtifactDataByName(artifactName)
	for i=1, artifacts.races do
		for k,v in pairs(artifacts[i].common) do
			if k == artifactName then
				artifacts[i].common[k] = nil
			end
		end
		for k,v in pairs(artifacts[i].rare) do
			if k == artifactName then
				artifacts[i].rare[k] = nil
			end
		end
	end
end

local function findItemLink(itemName)
	local r, _ = itemName
	for i = 1, artifacts.races do
		if artifactIDs[i].rare[itemName] then _, r = GetItemInfo(artifactIDs[i].rare[itemName]) end
		if artifactIDs[i].common[itemName] then _, r = GetItemInfo(artifactIDs[i].common[itemName]) end
		if artifactIDs[i].special[itemName] then _, r =  GetItemInfo(artifactIDs[i].special[itemName]) end
	end
	return r
end

local function getProgress(historyFrame)
	for i=1, GetNumArchaeologyRaces()-1 do -- That function returns 10, but includes that damn "Other" race... :P
		artifacts[i].progress.completedArtifacts = GetNumArtifactsByRace(i)

		-- Find out which artifacts have been completed, and how many times

		artifacts[i].progress.totalSolved = 0

		for j=1, GetNumArtifactsByRace(i) do
			local name, _, rarity, _, _, _, _, _, count = GetArtifactInfoByRace(i, j)
			if count > 0 then
				if linkedProjects[name] then name = linkedProjects[name] end
				if artifacts[i][rarity == 0 and "common" or "rare"][name] == nil then
					potentialLinks[#potentialLinks + 1] = name
				end
				artifacts[i][rarity == 0 and "common" or "rare"][name] = true
			end
			artifacts[i].progress.totalSolved = artifacts[i].progress.totalSolved + count
		end

		-- "Un-count" the artifact in progress if it has not yet been completed once

		local name, _, rarity = GetActiveArtifactByRace(i)
		if not artifacts[i][rarity == 0 and "common" or "rare"][name] then artifacts[i].progress.completedArtifacts = max(artifacts[i].progress.completedArtifacts - 1, 0) end

		-- Find the progress of the current artifact

		SetSelectedArtifact(i)
		local progress, _, maximum = GetArtifactProgress()
		artifacts[i].progress.currentProgress = progress
		artifacts[i].progress.currentMax = maximum
		local name, _, _, _, _, keystones = GetActiveArtifactByRace(i)
		if name == nil then -- This happens if the current race has not yet been opened.
			artifacts[i].progress.currentProgress = 0
			artifacts[i].progress.currentMax = 0
		end
		artifacts[i].progress.currentName = name
		artifacts[i].progress.numSlots = keystones or 0

		-- Count keystones

		local _, _, itemID = GetArchaeologyRaceInfo(i)
		local count = GetItemCount(itemID, false, false)
		artifacts[i].progress.numKeystones = count
	end
	if historyFrame then historyFrame:SetScript("OnEvent", getProgress) end
	getProgressCallback()
end

local specialFrame = CreateFrame("Frame")
specialFrame:RegisterEvent("TRADE_SKILL_SHOW")
local function specialEvent(self, event, ...)
	if event == "TRADE_SKILL_SHOW" then
		historyFrame:Hide()
		excavatinator()
		historyFrame:Show()
		local prof, prof2 = GetProfessions()
		if prof2 then
			local name, _, _, _, _, _, skillLine = GetProfessionInfo(prof2)
			if skillLine == 171 then prof = prof2 end
		end
		if prof then
			local name, _, _, _, _, _, skillLine = GetProfessionInfo(prof)
			local itemName = GetItemInfo(67538)
			if skillLine == 171 then
				local openName = GetTradeSkillLine()
				if name == openName and not IsTradeSkillLinked() then -- Alchemy is open.
					local vialName, skillName = GetSpellInfo(93328), ""
					for i=1, GetNumTradeSkills() do
						skillName = GetTradeSkillInfo(i)
						if vialName == skillName then
							-- Yes! Vial is present.
							artifacts[7].special[itemName] = 1
						end
						if artifacts[7].special[itemName] == -1 then
							-- Vial not found - register as incomplete.
							artifacts[7].special[itemName] = 0
						end
					end
				end
			else
				if itemName then -- Should be fixed, but best make sure it doesn't cause an error.
					artifacts[7].special[itemName] = 0
				end
			end
		end
	end
end
specialFrame:SetScript("OnEvent", specialEvent)

local function solveArtifact(raceID)
	SetSelectedArtifact(raceID)
	for i=1, min(artifacts[raceID].progress.numSlots, artifacts[raceID].progress.numKeystones) do
		SocketItemToArtifact()
	end
	SolveArtifact()
end

local leftLabel, rightLabel, leftString, rightString, listLength = {}, {}, {}, {}, 0

local function startCooldown(self, event, unit, spellName, spellRank, _, spellID)
	if unit == "player" and spellID == self.spellID then
		self:SetCooldown(GetTime(), self.cooldown)
	end
end

local function makeSecureButton(name, parent, texture, isItem, itemID, spellID, cooldownLength)
	local r = CreateFrame("Button", name, parent, "SecureActionButtonTemplate")
	r.dtexture = r:CreateTexture()
	r.dtexture:SetTexture(texture)
	r.dtexture:SetAllPoints()
	r.dtexture:SetDesaturated(true)
	r.ctexture = r:CreateTexture()
	r.ctexture:SetTexture("Interface/Buttons/UI-Quickslot-Depress")
	r.ctexture:SetAllPoints()
	r.ctexture:SetBlendMode("ADD")
	r.ctexture:SetDrawLayer("BORDER")
	r.htexture = r:CreateTexture()
	r.htexture:SetTexture("Interface/Buttons/ButtonHilight-Square")
	r.htexture:SetAllPoints()
	r.htexture:SetBlendMode("ADD")
	r.htexture:SetDrawLayer("HIGHLIGHT")
	r.ntexture = r:CreateTexture()
	r.ntexture:SetTexture(texture)
	r.ntexture:SetDrawLayer("BACKGROUND")
	r.ntexture:SetAllPoints()
	r:SetPushedTexture(r.ctexture)
	r:SetDisabledTexture(r.dtexture)
	r:SetWidth(32)
	r:SetHeight(32)
	r.label = r:CreateFontString()
	r.label:SetFontObject(NumberFontNormal)
	r.label:SetPoint("BOTTOMRIGHT")
	if isItem then
		r:SetAttribute("type", "item")
		r:SetAttribute("item", GetItemInfo(itemID))
	else
		r:SetAttribute("type", "spell")
		r:SetAttribute("spell", GetSpellInfo(spellID))
		r:SetAttribute("unit", "player")
	end
	r.cooldown = CreateFrame("Cooldown", name .. "Cooldown", r)
	r.cooldown:SetAllPoints(r)
	r.cooldown:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	r.cooldown.cooldown = cooldownLength
	r.cooldown.spellID = spellID
	r.cooldown:SetScript("OnEvent", startCooldown)
	return r
end

local function tooltipEnterFunc(self)
	if self.link then 
		GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
		GameTooltip:ClearLines()
		GameTooltip:SetHyperlink(self.link)
		GameTooltip:Show()
	end
end

local function tooltipLeaveFunc()
	GameTooltip:Hide()
end

local function linkClickFunc(self)
	if self.link then ChatEdit_InsertLink(self.link) end
end

function Excavatinator_setFrame(frame)
	local f = frame
	f:SetBackdrop({
		bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
		edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
		tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 5, right = 5, top = 5, bottom = 5 } } )
	for i=1, 30 do
		rightLabel[i] = makeLabel(f, 10, "[" .. i .. " In progress]")
		rightLabel[i]:SetPoint("TOPRIGHT", -8, - (12 * i) - 10)
		leftLabel[i] = CreateFrame("Button", "EXCLeftLabel"..i, f)
		leftLabel[i]:SetNormalFontObject('GameFontHighlightSmallLeft')
		leftLabel[i]:SetDisabledFontObject('GameFontHighlightSmallLeft')
		leftLabel[i]:SetWidth(200)
		leftLabel[i]:SetHeight(10)
		leftLabel[i]:SetPushedTextOffset(0, 0)
		leftLabel[i]:SetPoint("TOPLEFT", 8, - (12 * i) - 10)
		leftLabel[i]:SetScript("OnEnter", tooltipEnterFunc)
		leftLabel[i]:SetScript("OnLeave", tooltipLeaveFunc)
		leftLabel[i]:SetScript("OnClick", linkClickFunc)
	end
	f:SetPoint("CENTER")
end

function Excavatinator_UpdateScrollFrame()
	FauxScrollFrame_Update(ExcavatinatorListFrame,listLength,30,16, nil, nil, nil, nil, nil, nil, true)
	local line, lineoffset;
	for line=1, 30 do
		lineoffset = line + (FauxScrollFrame_GetOffset(ExcavatinatorListFrame) or 0)
		local link, _ = ""
		if ExcavatinatorListFrame.race == 0 then
			link = findItemLink(leftString[lineoffset])
		else
			_, link = GetItemInfo(artifactIDs[ExcavatinatorListFrame.race].common[leftString[lineoffset]] or artifactIDs[ExcavatinatorListFrame.race].rare[leftString[lineoffset]] or (artifactIDs[ExcavatinatorListFrame.race].special and artifactIDs[ExcavatinatorListFrame.race].special[leftString[lineoffset]] or nil) or 0)
		end
		leftLabel[line]:SetText(link or leftString[lineoffset])
		if not link then leftLabel[line]:Disable() else leftLabel[line]:Enable() end
		leftLabel[line]:Show()
		rightLabel[line]:SetText(rightString[lineoffset])
		rightLabel[line]:Show()
		if link then
			leftLabel[line].link = link
		else
			leftLabel[line].link = nil
		end
	end
end

local function populateScrollFrame(raceID)
	local n = 2
	leftString[1] = "|cFFFFFF00RARE ARTIFACTS|r"
	rightString[1] = ""
	for k,v in pairs(artifacts[raceID].rare) do
		leftString[n] = k or "Item not found"
		rightString[n] = v and "|cFF00FF00[Completed]|r" or "|cFFFF0000[Incomplete]|r"
		n = n + 1
	end
	leftString[n] = ""
	rightString[n] = ""
	leftString[n+1] = "|cFFFFFF00COMMON ARTIFACTS|r"
	rightString[n+1] = ""
	n = n + 2
	for k,v in pairs(artifacts[raceID].common) do
		leftString[n] = k or "Item not found"
		rightString[n] = v and "|cFF00FF00[Completed]|r" or "|cFFFF0000[Incomplete]|r"
		n = n + 1
	end
	if artifacts[raceID].special and (artifacts[raceID].special.n > 0) then
		leftString[n] = ""
		rightString[n] = ""
		leftString[n+1] = "|cFFFFFF00OTHER|r"
		rightString[n+1] = ""
		n = n + 2
		for k,v in pairs(artifacts[raceID].special) do
			if k ~= "n" then
				leftString[n] = k
				rightString[n] = v == -1 and "|c00FFFF00[Unknown]|r" or (v == 0 and "|cFFFF0000[Incomplete]|r" or "|cFF00FF00[Completed]|r")
				n = n + 1
			end
		end
	end
	listLength = n - 1
	while leftString[n] do
		leftString[n] = nil
		rightString[n] = nil -- Safe to do because the lists are always the same length.
		n = n + 1
	end
	ExcavatinatorListFrame.race = raceID
	ExcavatinatorListFrame:Hide()
	FauxScrollFrame_SetOffset(ExcavatinatorListFrame, 0)
	FauxScrollFrame_Update(ExcavatinatorListFrame, 1, 1, 16)
	if listLength <= 30 then
		ExcavatinatorListFrameScrollBar:Hide()
	else
		ExcavatinatorListFrameScrollBar:Show()
	end
	ExcavatinatorListFrame:Show()
end

local function populateScrollFrameByMissing(includeRare, includeCommon)
	local n = 1
	if includeRare then
		leftString[n] = "|cFFFFFF00MISSING RARE ARTIFACTS|r"
		rightString[n] = ""
		n = n + 1
		for i = 1, artifacts.races do
			for k, v in pairs(artifacts[i].rare) do
				if not v then
					leftString[n] = k or "Item not found"
					rightString[n] = GetArchaeologyRaceInfo(i)
					n = n + 1
				end
			end
		end
	end
	if includeCommon then
		if n > 1 then
			leftString[n] = ""
			rightString[n] = ""
			n = n + 1
		end
		leftString[n] = "|cFFFFFF00MISSING COMMON ARTIFACTS|r"
		rightString[n] = ""
		n = n + 1
		for i = 1, artifacts.races do
			for k, v in pairs(artifacts[i].common) do
				if not v then
					leftString[n] = k or "Item not found"
					rightString[n] = GetArchaeologyRaceInfo(i)
					n = n + 1
				end
			end
		end
	end
	listLength = n - 1
	while leftString[n] do
		leftString[n] = nil
		rightString[n] = nil -- Safe to do because the lists are always the same length.
		n = n + 1
	end
	ExcavatinatorListFrame.race = 0
	ExcavatinatorListFrame:Hide()
	FauxScrollFrame_SetOffset(ExcavatinatorListFrame, 0)
	FauxScrollFrame_Update(ExcavatinatorListFrame, 1, 1, 16)
	if listLength <= 30 then
		ExcavatinatorListFrameScrollBar:Hide()
	else
		ExcavatinatorListFrameScrollBar:Show()
	end
	ExcavatinatorListFrame:Show()
end

SLASH_EXCAVATINATOR1 = '/arch'

local updateFrame = CreateFrame("Frame")
updateFrame:RegisterEvent("ARTIFACT_COMPLETE")
updateFrame:RegisterEvent("CHAT_MSG_CURRENCY")
updateFrame:RegisterEvent("BAG_UPDATE")
updateFrame:RegisterEvent("ZONE_CHANGED_NEW_AREA")
updateFrame:RegisterEvent("CRITERIA_UPDATE")

local function expand()
	raceFrame:SetHeight(252)
	for i=1, artifacts.races do
		raceFrame.button[i]:SetPoint("TOPLEFT", 12, -12-20*i)
		raceFrame.button[i]:Show()
		raceFrame.progressBar[i]:SetPoint("TOPLEFT", 120, -17-20*i)
		raceFrame.progressBar[i]:Show()
		raceFrame.currentBar[i]:SetPoint("TOPLEFT", 300, -17-20*i)
		raceFrame.currentBar[i]:Show()
		raceFrame.keystones[i]:SetPoint("TOPLEFT", 500, -16-20*i)
		raceFrame.keystones[i]:Show()
	end
	raceFrame.expand:Hide()
	raceFrame.collapse:Show()
	raceFrame.isCollapsed = false
end

local function collapse()
	-- Get current continent
	SetMapToCurrentZone()
	local continent = GetCurrentMapContinent()
	if continent < 1 or continent > 4 then continent = 0 end -- Failsafe

	raceFrame:SetHeight(75 + 20 * #continents[continent])
	for i=1, artifacts.races do
		raceFrame.button[i]:Hide()
		raceFrame.progressBar[i]:Hide()
		raceFrame.currentBar[i]:Hide()
		raceFrame.keystones[i]:Hide()
	end
	for i=1, #continents[continent] do
		local j = continents[continent][i]
		raceFrame.button[j]:SetPoint("TOPLEFT", 12, -12-20*i)
		raceFrame.button[j]:Show()
		raceFrame.progressBar[j]:SetPoint("TOPLEFT", 120, -17-20*i)
		raceFrame.progressBar[j]:Show()
		raceFrame.currentBar[j]:SetPoint("TOPLEFT", 300, -17-20*i)
		raceFrame.currentBar[j]:Show()
		raceFrame.keystones[j]:SetPoint("TOPLEFT", 500, -16-20*i)
		raceFrame.keystones[j]:Show()
	end
	raceFrame.collapse:Hide()
	raceFrame.expand:Show()
	raceFrame.isCollapsed = true
end

local function surveyButton_onEvent(self, event, unit, spellname, spellrank, something, spellID)
	if IsSpellKnown(80451) then
		raceFrame.survey:Enable()
	else
		raceFrame.survey:Disable()
	end
end

local achievementIDs = { 5470, 4856, 5511, 4858, 5193, 4859, 5191, 5192, 5301, 4923 }
local achievementItems = {
	{},
	{}, -- First three don't have any.
	{},
	{ 0, 64367, 64487, 64444, 64382, 64475, 64420, 64348, { 1, 1, 2, 4, 5, 6, 8 } }, -- Seven Scepters
	{ 1, 64337, 64362, 64366, 64483 }, -- Blue Streak
	{ 1, 64368, 63414, 64343, 64344, 64371, 64485, 64484 }, -- Kings Under the Mountain
	{ 4, 64357, 63407, 64356, 64379, 64381, 64648 }, -- Tragedy in Three Acts
	{ 6, 64389, 64417, 64418, 64419 }, -- The Harder they Fall
	{ 7, 64653, 64654, 64655, 64656, 64658, 64657 }, -- The Boy Who Would be King
	{},
}
local requiredSkill = { 0, 300, 0, 0, 375, 300, 450, 0, 375 }

local function updateAchievements()
	if linksFrame then linksFrame:Hide() end
	if optionsFrame then optionsFrame:Hide() end
	-- Make sure all buttons and bars are shown
	for i = 1, 10 do
		achieveFrame.button[i]:Show()
		achieveFrame.progress[i]:Show()
	end

	-- Hide everything else
	achieveFrame.header:Hide()
	for i=1, 9 do
		achieveFrame.line[i]:Hide()
		achieveFrame.rightLine[i]:Hide()
		achieveFrame.line[i].item = 0
	end

	achieveFrame.button[1]:Enable()
	achieveFrame.ID = 0

	-- 1. Diggerest
	local _, name = GetAchievementInfo(5470)
	achieveFrame.button[1]:SetText(name)
	local _, _, _, progress = GetAchievementCriteriaInfo(5470, 1)
	if progress == 0 then
		achieveFrame.progress[1].bar:Hide()
	else
		achieveFrame.progress[1].bar:Show()
		achieveFrame.progress[1].bar:SetWidth(80 * (progress / 100))
	end
	achieveFrame.progress[1].progress = progress
	achieveFrame.progress[1].label:SetText(progress .. "/100")

	-- 2. It Belongs in a Museum!
	_, name = GetAchievementInfo(4856)
	achieveFrame.button[2]:SetText(name)
	_, _, _, progress = GetAchievementCriteriaInfo(4856, 1)
	if progress == 0 then
		achieveFrame.progress[2].bar:Hide()
	else
		achieveFrame.progress[2].bar:Show()
		achieveFrame.progress[2].bar:SetWidth(80 * (progress / 20))
	end
	achieveFrame.progress[2].progress = progress
	achieveFrame.progress[2].label:SetText(progress .. "/20")

	-- 3. It's Always in the Last Place You Look
	_, name = GetAchievementInfo(5511)
	achieveFrame.button[3]:SetText(name)
	progress = 0
	achieveFrame.progress[3].race = {}
	for i=1, GetAchievementNumCriteria(5511) do
		if artifacts[i].progress.completedArtifacts > 0 then
			achieveFrame.progress[3].race[i] = true
			progress = progress + 1
		else
			achieveFrame.progress[3].race[i] = false
		end
	end
	if progress == 0 then
		achieveFrame.progress[3].bar:Hide()
	else
		achieveFrame.progress[3].bar:Show()
		achieveFrame.progress[3].bar:SetWidth(80 * (progress / 9))
	end
	achieveFrame.progress[3].progress = progress
	achieveFrame.progress[3].label:SetText(progress .. "/9")

	-- 4. Seven Scepters
	_, name = GetAchievementInfo(4858)
	achieveFrame.button[4]:SetText(name)
	progress = 0
	for i = 2, 8 do
		if artifacts[achievementItems[4][9][i-1]].common[GetItemInfo(achievementItems[4][i])] then progress = progress + 1 end
	end
	if progress == 0 then
		achieveFrame.progress[4].bar:Hide()
	else
		achieveFrame.progress[4].bar:Show()
		achieveFrame.progress[4].bar:SetWidth(80 * (progress / 7))
	end
	achieveFrame.progress[4].progress = progress
	achieveFrame.progress[4].label:SetText(progress .. "/7")

	for i=5, 9 do
		-- 5. Blue Streak
		-- 6. Kings Under the Mountain
		-- 7. Tragedy in Three Acts
		-- 8. The Harder they Fall
		-- 9. The Boy Who Would be King
		_, name = GetAchievementInfo(achievementIDs[i])
		achieveFrame.button[i]:SetText(name)
		progress = 0
		for j = 2, #achievementItems[i] do
			if artifacts[achievementItems[i][1]].common[GetItemInfo(achievementItems[i][j])] then progress = progress + 1 end
		end
		if progress == 0 then
			achieveFrame.progress[i].bar:Hide()
		else
			achieveFrame.progress[i].bar:Show()
			achieveFrame.progress[i].bar:SetWidth(80 * (progress / (#achievementItems[i] - 1)))
		end
		achieveFrame.progress[i].progress = progress
		achieveFrame.progress[i].label:SetText(progress .. "/" .. (#achievementItems[i] - 1))
	end

	-- 10. Illustrious Grand Master in Archaeology
	_, name = GetAchievementInfo(4923)
	achieveFrame.button[10]:SetText(name)
	local _, _, archIndex = GetProfessions()
	if archIndex then _, _, progress = GetProfessionInfo(archIndex) else progress = 0 end
	if progress == 0 then
		achieveFrame.progress[10].bar:Hide()
	else
		achieveFrame.progress[10].bar:Show()
		achieveFrame.progress[10].bar:SetWidth(80 * (progress / 525))
	end
	achieveFrame.progress[10].progress = progress
	achieveFrame.progress[10].label:SetText(progress .. "/525")

	achieveFrame:Show()
end

local function expandAchievement(ID)
	achieveFrame.ID = ID
	achieveFrame.button[1]:Disable()

	for i=2, 10 do
		achieveFrame.button[i]:Hide()
		achieveFrame.progress[i]:Hide()
	end
	for i=1, 9 do
		achieveFrame.line[i]:Hide()
		achieveFrame.rightLine[i]:Hide()
		achieveFrame.line[i].item = 0
	end
	achieveFrame.header:Show()

	achieveFrame.button[1]:SetText(achieveFrame.button[ID]:GetText())
	if achieveFrame.progress[ID].bar:IsShown() then
		achieveFrame.progress[1].bar:Show()
		achieveFrame.progress[1].bar:SetWidth(achieveFrame.progress[ID].bar:GetWidth())
	else
		achieveFrame.progress[1].bar:Hide()
	end
	achieveFrame.progress[1].label:SetText(achieveFrame.progress[ID].label:GetText())

	local n = 1

	if ID == 1 then -- Diggerest
		if achieveFrame.progress[1].progress == 100 then
			achieveFrame.header:SetText("|cFF00FF00You have completed this achievement!|r")
		else
			achieveFrame.header:SetText("You need " .. (100 - achieveFrame.progress[1].progress) .. " more artifacts.")
		end
		achieveFrame.line[1]:Show()
		achieveFrame.line[1]:SetText("Click here for a list of undiscovered common artifacts.")
		achieveFrame.line[1]:Enable()
		achieveFrame.line[1]:SetScript("OnClick", function() populateScrollFrameByMissing(false, true) end)
	elseif ID == 2 then -- It Belongs in a Museum!
		if achieveFrame.progress[2].progress == 20 then
			achieveFrame.header:SetText("|cFF00FF00You have completed this achievement!|r")
		else
			achieveFrame.header:SetText("You need " .. (20 - achieveFrame.progress[2].progress) .. " more artifacts.")
		end
		achieveFrame.line[1]:Show()
		achieveFrame.line[1]:SetText("Click here for a list of undiscovered rare artifacts.")
		achieveFrame.line[1]:Enable()
		achieveFrame.line[1]:SetScript("OnClick", function() populateScrollFrameByMissing(true, false) end)
	elseif ID == 3 then -- It's Always in the Last Place You Look
		if achieveFrame.progress[3].progress == 9 then
			achieveFrame.header:SetText("|cFF00FF00You have completed this achievement!|r")
		else
			achieveFrame.header:SetText("You still need artifacts from " .. (achieveFrame.progress[3].progress == 8 and "this race:" or "these " .. (9 - achieveFrame.progress[3].progress) .. " races:"))
			n = 1
			for i = 1, artifacts.races do
				if not achieveFrame.progress[3].race[i] then
					achieveFrame.line[n]:Show()
					achieveFrame.line[n]:SetText(GetArchaeologyRaceInfo(i))
					achieveFrame.line[n]:Disable()
					if achieveFrame.progress[10].progress < requiredSkill[i] then
						achieveFrame.rightLine[n]:Show()
						achieveFrame.rightLine[n]:SetText("|c00FF0000Requires " .. requiredSkill[i] .. " skill points.|r")
					end
					n = n + 1
				end
			end
		end
	elseif ID == 4 then -- Seven Scepters
		if achieveFrame.progress[4].progress == 7 then
			achieveFrame.header:SetText("|cFF00FF00You have completed this achievement!|r")
		else
			achieveFrame.header:SetText("You still need the following " .. 7 - achieveFrame.progress[4].progress .. " artifacts:")
			n = 1
			local name, link;
			for i = 2, 8 do
				name, link = GetItemInfo(achievementItems[4][i])
				if not artifacts[achievementItems[4][9][i-1]].common[name] then
					achieveFrame.line[n]:Show()
					achieveFrame.line[n]:SetText(link)
					achieveFrame.line[n]:Enable()
					achieveFrame.line[n].item = achievementItems[4][i]
					achieveFrame.rightLine[n]:Show()
					achieveFrame.rightLine[n]:SetText(GetArchaeologyRaceInfo(achievementItems[4][9][i-1]))
					n = n + 1
				end
			end
		end
	elseif ID > 4 and ID < 10 then
		if achieveFrame.progress[ID].progress == #achievementItems[ID] - 1 then
			achieveFrame.header:SetText("|cFF00FF00You have completed this achievement!|r")
		else
			achieveFrame.header:SetText("You still need the following " .. #achievementItems[ID] - 1 - achieveFrame.progress[ID].progress .. " " .. strlower(GetArchaeologyRaceInfo(achievementItems[ID][1])) .. " artifacts:")
			n = 1
			local name, link;
			for i = 2, #achievementItems[ID] do
				name, link = GetItemInfo(achievementItems[ID][i])
				if not artifacts[achievementItems[ID][1]].common[name] then
					achieveFrame.line[n]:Show()
					achieveFrame.line[n]:SetText(link)
					achieveFrame.line[n]:Enable()
					achieveFrame.line[n].item = achievementItems[ID][i]
					n = n + 1
				end
			end
		end
	elseif ID == 10 then
		if achieveFrame.progress[10].progress == 525 then
			achieveFrame.header:SetText("|cFF00FF00You have completed this achievement!|r")
		else
			achieveFrame.header:SetText("You still need " .. 525 - achieveFrame.progress[10].progress .. " skill points in archaeology.")
		end
	end
end

local TRAVELFRAME_MAXHEIGHT = 8

function openTravelFrame()
	if itemsFrame then itemsFrame:Hide() end
	-- Create travel frame if it does not exist, then show it
	if not travelFrame then
		travelFrame = CreateFrame("Frame", "ExcavatinatorTravelFrame", raceFrame)
		travelFrame:SetBackdrop( {
			bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
			edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
			tile = true, tileSize = 16, edgeSize = 16,
			insets = { left = 5, right = 5, top = 5, bottom = 5 } } )
		travelFrame:SetPoint("TOPRIGHT", raceFrame, "TOPLEFT")
		travelFrame:SetWidth(48)
		travelFrame:SetHeight(48)
		travelFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
		travelFrame:SetScript("OnEvent", function(self, event, unit) if unit == "player" then self:Hide() end end)
		-- Create buttons
		travelFrame.itemButtons = {}
		travelFrame.spellButtons = {}
		for i=1, #travel.items do
			local texture = GetItemIcon(travel.items[i][1])
			travelFrame.itemButtons[i] = makeSecureButton("ExcavatinatorTravelItemButton"..i, travelFrame, texture, true, travel.items[i][1], travel.items[i][2], travel.items[i][3])
			travelFrame.itemButtons[i].itemID = travel.items[i][1]
			travelFrame.itemButtons[i].consumed = travel.items[i][4]
			local _, link = GetItemInfo(travel.items[i][1])
			travelFrame.itemButtons[i]:SetScript("OnEnter", function()
				GameTooltip:SetOwner(travelFrame.itemButtons[i], "ANCHOR_CURSOR")
				GameTooltip:ClearLines()
				GameTooltip:SetHyperlink(link)
			end)
			travelFrame.itemButtons[i]:SetScript("OnLeave", function() GameTooltip:Hide() end)
		end
		for i=1, #travel.spells do
			local _, _, texture = GetSpellInfo(travel.spells[i][1])
			travelFrame.spellButtons[i] = makeSecureButton("ExcavatinatorTravelSpellButton"..i, travelFrame, texture, false, nil, travel.spells[i][1], travel.spells[i][2])
			travelFrame.spellButtons[i].spellID = travel.spells[i][1]
			travelFrame.spellButtons[i].reagent = travel.spells[i][3]
			local link = GetSpellLink(travel.spells[i][1])
			travelFrame.spellButtons[i]:SetScript("OnEnter", function()
				GameTooltip:SetOwner(travelFrame.spellButtons[i], "ANCHOR_CURSOR")
				GameTooltip:ClearLines()
				GameTooltip:SetHyperlink(link)
			end)
			travelFrame.spellButtons[i]:SetScript("OnLeave", function() GameTooltip:Hide() end)
		end
	elseif travelFrame:IsShown() then
		travelFrame:Hide()
	else
		travelFrame:Show()
	end

	-- Update buttons
	local n = 0
	for i=1, #travelFrame.itemButtons do
		local count = GetItemCount(travelFrame.itemButtons[i].itemID)
		if count > 0 then
			travelFrame.itemButtons[i]:Show()
			travelFrame.itemButtons[i]:SetPoint("TOPRIGHT", -8 - (36 * math.floor(n / TRAVELFRAME_MAXHEIGHT)), -8 -(36 * (n % TRAVELFRAME_MAXHEIGHT)))
			n = n + 1
			local start, duration = GetItemCooldown(travelFrame.itemButtons[i].itemID)
			if duration > 0 then
				travelFrame.itemButtons[i].cooldown:SetCooldown(start, duration)
			end
			if travelFrame.itemButtons[i].consumed then
				travelFrame.itemButtons[i].label:Show()
				travelFrame.itemButtons[i].label:SetText(count)
			else
				travelFrame.itemButtons[i].label:Hide()
			end
		else
			travelFrame.itemButtons[i]:Hide()
		end
	end
	for i=1, #travelFrame.spellButtons do
		if IsSpellKnown(travelFrame.spellButtons[i].spellID) then
			travelFrame.spellButtons[i]:Show()
			travelFrame.spellButtons[i]:SetPoint("TOPRIGHT", -8 - (36 * math.floor(n / TRAVELFRAME_MAXHEIGHT)), -8 -(36 * (n % TRAVELFRAME_MAXHEIGHT)))
			n = n + 1
			local start, duration = GetSpellCooldown(travelFrame.spellButtons[i].spellID)
			if duration > 0 then
				travelFrame.spellButtons[i].cooldown:SetCooldown(start, duration)
			end
			if travelFrame.spellButtons[i].reagent ~= 0 then
				local count = GetItemCount(travelFrame.spellButtons[i].reagent)
				if count == 0 then
					travelFrame.spellButtons[i]:Disable()
				else
					travelFrame.spellButtons[i]:Enable()
				end
			travelFrame.spellButtons[i].label:Show()
			travelFrame.spellButtons[i].label:SetText(count)
			end
		else
			travelFrame.spellButtons[i]:Hide()
		end
	end

	-- Resize the frame
	travelFrame:SetWidth(max(36 * math.floor(((n - 1) / TRAVELFRAME_MAXHEIGHT) + 1) + 12, 48))
	travelFrame:SetHeight(max(36 * min(n, TRAVELFRAME_MAXHEIGHT) + 12, 48))
end

local function orderedLinksList()
	local r = {}
	for k, v in pairs(linkedProjects) do
		r[#r+1] = { k, v }
	end
	return r
end

local function orderedRemainingArtifactsList()
	local r = {}
	for i = 1, artifacts.races do
		for k,v in pairs(artifacts[i].common) do
			if not v then
				if not linkedProjectsByItemName[k] then r[#r+1] = k end
			end
		end
		for k,v in pairs(artifacts[i].rare) do
			if not v then
				if not linkedProjectsByItemName[k] then r[#r+1] = k end
			end
		end
	end
	return r
end

local function updateLinksDeleteButtons(links, offset)
	for i = 1, min(#links - offset, 8) do
		linksFrame.modify.delete[i]:Enable()
		linksFrame.modify.delete[i].ID = i + offset
		linksFrame.modify.artifactLabel[i]:SetText(links[i + offset][1])
		linksFrame.modify.itemLabel[i]:SetText(links[i + offset][2])
	end
	for i = #links + 1 - offset, 8 do
		linksFrame.modify.delete[i]:Disable()
		linksFrame.modify.artifactLabel[i]:SetText("")
		linksFrame.modify.itemLabel[i]:SetText("")
	end
	if offset > 0 then linksFrame.modify.up:Enable() else linksFrame.modify.up:Disable() end
	if #links - offset > 8 then linksFrame.modify.down:Enable() else linksFrame.modify.down:Disable() end
end

local function updateLinksCreateLeftButtons(offset)
	for i = 1, min(#potentialLinks - offset, 12) do
		linksFrame.create.left.button[i]:Enable()
		linksFrame.create.left.button[i].ID = i + offset
		linksFrame.create.left.button[i]:SetText(potentialLinks[i + offset])
	end
	for i = #potentialLinks + 1 - offset, 12 do
		linksFrame.create.left.button[i]:Disable()
		linksFrame.create.left.button[i]:SetText("")
	end
	if offset > 0 then linksFrame.create.left.up:Enable() else linksFrame.create.left.up:Disable() end
	if #potentialLinks - offset > 12 then linksFrame.create.left.down:Enable() else linksFrame.create.left.down:Disable() end
end

local function updateLinksCreateRightButtons(itemList, offset)
	for i = 1, min(#itemList - offset, 12) do
		linksFrame.create.right.button[i]:Enable()
		linksFrame.create.right.button[i].ID = i + offset
		linksFrame.create.right.button[i]:SetText(itemList[i + offset])
	end
	for i = #itemList + 1 - offset, 12 do
		linksFrame.create.right.button[i]:Disable()
		linksFrame.create.right.button[i]:SetText("")
	end
	if offset > 0 then linksFrame.create.right.up:Enable() else linksFrame.create.right.up:Disable() end
	if #itemList - offset > 12 then linksFrame.create.right.down:Enable() else linksFrame.create.right.down:Disable() end
end

local function createLink(projectID, itemID)
	linkedProjects[potentialLinks[projectID]] = linksFrame.items[itemID]
	linkedProjectsByItemName[linksFrame.items[itemID]] = potentialLinks[projectID]
	linksFrame.links[#linksFrame.links+1] = { potentialLinks[projectID], linksFrame.items[itemID] }
	deleteArtifactDataByName(potentialLinks[projectID])
	table.remove(potentialLinks, projectID)
	table.remove(linksFrame.items, itemID)
	linksFrame.create.left.ID = 0
	linksFrame.create.left.label:SetText("|cFFFFFF00Project:|r <none selected>")
	linksFrame.create.right.ID = 0
	linksFrame.create.right.label:SetText("|cFFFFFF00Item:|r <none selected>")
	linksFrame.create.button:Disable()
	updateLinksDeleteButtons(linksFrame.links, linksFrame.modify.offset)
end

local function deleteLink(self)
	linkedProjects[linksFrame.links[self.ID][1]] = nil
	linkedProjectsByItemName[linksFrame.links[self.ID][2]] = nil
	linksFrame.items[#linksFrame.items+1] = linksFrame.links[self.ID][2]
	potentialLinks[#potentialLinks+1] = linksFrame.links[self.ID][1]
	table.remove(linksFrame.links, self.ID)
	updateLinksCreateLeftButtons(linksFrame.create.left.offset)
	updateLinksCreateRightButtons(linksFrame.items, linksFrame.create.right.offset)
end

local function openLinksFrame()
	optionsFrame:Hide()
	if not linksFrame then
		linksFrame = CreateFrame("Frame", "ExcavatinatorLinksFrame", raceFrame)
		linksFrame:SetBackdrop( {
			bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
			edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
			tile = true, tileSize = 16, edgeSize = 16,
			insets = { left = 5, right = 5, top = 5, bottom = 5 } } )
		linksFrame:SetWidth(550)
		linksFrame:SetHeight(220)
		linksFrame:SetPoint("TOPLEFT", raceFrame, "BOTTOMLEFT")

		linksFrame.close = CreateFrame("Button", "ExcavatinatorLinksFrameCloseButton", linksFrame, "UIPanelCloseButton")
		linksFrame.close:SetPoint("TOPRIGHT", -8, -8)

		linksFrame.confirm = CreateFrame("Frame", "ExcavatinatorLinksConfirmFrame", linksFrame)
		linksFrame.confirm:SetBackdrop( {
			bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
			edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
			tile = true, tileSize = 16, edgeSize = 16,
			insets = { left = 5, right = 5, top = 5, bottom = 5 } } )
		linksFrame.confirm:SetWidth(700)
		linksFrame.confirm:SetHeight(100)
		linksFrame.confirm:SetFrameStrata("HIGH")
		linksFrame.confirm:EnableMouse(true)
		linksFrame.confirm:SetPoint("CENTER", UIParent, "CENTER")
		linksFrame.confirm.label = makeLabel(linksFrame.confirm, 16, "Label")
		linksFrame.confirm.label:SetWidth(650)
		linksFrame.confirm.label:SetPoint("TOP", 0, -16)
		linksFrame.confirm.ok = CreateFrame("Button", "ExcavatinatorLinksConfirmOK", linksFrame.confirm, "OptionsButtonTemplate")
		linksFrame.confirm.ok:SetPoint("BOTTOM", -50, 16)
		linksFrame.confirm.ok:SetText("OK")
		linksFrame.confirm.cancel = CreateFrame("Button", "ExcavatinatorLinksConfirmCancel", linksFrame.confirm, "OptionsButtonTemplate")
		linksFrame.confirm.cancel:SetPoint("BOTTOM", 50, 16)
		linksFrame.confirm.cancel:SetText("Cancel")
		linksFrame.confirm.self = linksFrame.confirm
		linksFrame.confirm.cancelFunc = function()
			linksFrame.confirm:Hide()
		end
		linksFrame.confirm.okdeletefunc = function()
			deleteLink(linksFrame.confirm.self)
			updateLinksDeleteButtons(linksFrame.links, linksFrame.modify.offset)
			linksFrame.confirm:Hide()
		end
		linksFrame.confirm.okcreatefunc = function()
			createLink(linksFrame.confirm.self.projectID, linksFrame.confirm.self.itemID)
			updateLinksCreateLeftButtons(linksFrame.create.left.offset)
			updateLinksCreateRightButtons(linksFrame.items, linksFrame.create.right.offset)
			linksFrame.confirm:Hide()
		end
		linksFrame.confirm.ok:SetScript("OnClick", linksFrame.confirm.cancelFunc)
		linksFrame.confirm.cancel:SetScript("OnClick", linksFrame.confirm.cancelFunc)
		linksFrame.confirm:Hide()

		linksFrame.modify = CreateFrame("Frame", "ExcavatinatorLinksModifyFrame", linksFrame)
		linksFrame.modify:SetAllPoints()
		linksFrame.modify.switch = CreateFrame("Button", "ExcavatinatorSwitchButton1", linksFrame.modify)
		linksFrame.modify.switch:SetNormalTexture("Interface/Buttons/UI-RotationRight-Button-Up")
		linksFrame.modify.switch:SetPushedTexture("Interface/Buttons/UI-RotationRight-Button-Down")
		linksFrame.modify.switch:SetWidth(32)
		linksFrame.modify.switch:SetHeight(32)
		linksFrame.modify.switch:SetPoint("BOTTOMRIGHT", -8, 8)
		linksFrame.modify.switch:SetScript("OnClick", function()
			linksFrame.modify:Hide()
			linksFrame.create:Show()
		end)
		linksFrame.modify.header1 = makeLabel(linksFrame.modify, 12, "|cFFFFFF00Artifact|r")
		linksFrame.modify.header1:SetPoint("TOPLEFT", 66, -8)
		linksFrame.modify.header2 = makeLabel(linksFrame.modify, 12, "|cFFFFFF00Item|r")
		linksFrame.modify.header2:SetPoint("TOPLEFT", 316, -8)
		linksFrame.modify.up = CreateFrame("Button", "ExcavatinatorLinksModifyUpButton", linksFrame.modify)
		linksFrame.modify.up:SetNormalTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Up")
		linksFrame.modify.up:SetPushedTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Down")
		linksFrame.modify.up:SetDisabledTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Disabled")
		linksFrame.modify.up:SetHighlightTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Highlight")
		linksFrame.modify.up:SetWidth(32)
		linksFrame.modify.up:SetHeight(32)
		linksFrame.modify.up:SetPoint("TOPLEFT", 8, -20)
		linksFrame.modify.up:SetScript("OnClick", function()
			linksFrame.modify.offset = max(linksFrame.modify.offset - 8, 0)
			updateLinksDeleteButtons(linksFrame.links, linksFrame.modify.offset)
		end)
		linksFrame.modify.down = CreateFrame("Button", "ExcavatinatorLinksModifyDownButton", linksFrame.modify)
		linksFrame.modify.down:SetNormalTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Up")
		linksFrame.modify.down:SetPushedTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Down")
		linksFrame.modify.down:SetDisabledTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Disabled")
		linksFrame.modify.down:SetHighlightTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Highlight")
		linksFrame.modify.down:SetWidth(32)
		linksFrame.modify.down:SetHeight(32)
		linksFrame.modify.down:SetPoint("BOTTOMLEFT", 8, 8)
		linksFrame.modify.down:SetScript("OnClick", function()
			linksFrame.modify.offset = linksFrame.modify.offset + 8
			updateLinksDeleteButtons(linksFrame.links, linksFrame.modify.offset)
		end)
		linksFrame.modify.delete = {}
		linksFrame.modify.artifactLabel = {}
		linksFrame.modify.itemLabel = {}
		linksFrame.modify.deleteButtonFunc = function(self)
			linksFrame.confirm.ok:SetScript("OnClick", linksFrame.confirm.okdeletefunc)
			linksFrame.confirm.self = self
			linksFrame.confirm.label:SetText("Delete the link between \"" .. linksFrame.links[self.ID][1] .. "\" and \"" .. linksFrame.links[self.ID][2] .. "\"?")
			linksFrame.confirm:Show()
		end
		for i=1, 8 do
			linksFrame.modify.delete[i] = CreateFrame("Button", "ExcavatinatorLinksModifyFrameDeleteButton" .. i, linksFrame.modify, "UIPanelCloseButton")
			linksFrame.modify.delete[i]:SetPoint("TOPLEFT", 36, 8 - (24 * i))
			linksFrame.modify.delete[i]:SetHitRectInsets(0, 0, 0, 0)
			linksFrame.modify.delete[i]:SetScript("OnClick", linksFrame.modify.deleteButtonFunc)
			linksFrame.modify.artifactLabel[i] = makeLabel(linksFrame.modify, 10, "Left label " .. i)
			linksFrame.modify.artifactLabel[i]:SetPoint("TOPLEFT", 66, - 4 - (24 * i))
			linksFrame.modify.itemLabel[i] = makeLabel(linksFrame.modify, 10, "Right label " .. i)
			linksFrame.modify.itemLabel[i]:SetPoint("TOPLEFT", 316, - 4 - (24 * i))
		end
		linksFrame.modify.offset = 0

		linksFrame.create = CreateFrame("Frame", "ExcavatinatorLinksCreateFrame", linksFrame)
		linksFrame.create:SetAllPoints()
		linksFrame.create.switch = CreateFrame("Button", "ExcavatinatorSwitchButton2", linksFrame.create)
		linksFrame.create.switch:SetNormalTexture("Interface/Buttons/UI-RotationRight-Button-Up")
		linksFrame.create.switch:SetPushedTexture("Interface/Buttons/UI-RotationRight-Button-Down")
		linksFrame.create.switch:SetWidth(32)
		linksFrame.create.switch:SetHeight(32)
		linksFrame.create.switch:SetPoint("BOTTOMRIGHT", -8, 8)
		linksFrame.create.switch:SetScript("OnClick", function()
			linksFrame.create:Hide()
			linksFrame.modify:Show()
		end)
		linksFrame.create.left = CreateFrame("Frame", "ExcavatinatorLinksCreateLeftFrame", linksFrame.create)
		linksFrame.create.left:SetWidth(250)
		linksFrame.create.left:SetHeight(180)
		linksFrame.create.left:SetPoint("TOPLEFT")
		linksFrame.create.left.up = CreateFrame("Button", "ExcavatinatorLinksCreateUpButton", linksFrame.create.left)
		linksFrame.create.left.up:SetNormalTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Up")
		linksFrame.create.left.up:SetPushedTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Down")
		linksFrame.create.left.up:SetDisabledTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Disabled")
		linksFrame.create.left.up:SetHighlightTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Highlight")
		linksFrame.create.left.up:SetWidth(32)
		linksFrame.create.left.up:SetHeight(32)
		linksFrame.create.left.up:SetPoint("TOPLEFT", 8, -20)
		linksFrame.create.left.up:SetScript("OnClick", function()
			linksFrame.create.left.offset = max(linksFrame.create.left.offset - 12, 0)
			updateLinksCreateLeftButtons(linksFrame.create.left.offset)
		end)
		linksFrame.create.left.down = CreateFrame("Button", "ExcavatinatorLinksCreateDownButton", linksFrame.create.left)
		linksFrame.create.left.down:SetNormalTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Up")
		linksFrame.create.left.down:SetPushedTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Down")
		linksFrame.create.left.down:SetDisabledTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Disabled")
		linksFrame.create.left.down:SetHighlightTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Highlight")
		linksFrame.create.left.down:SetWidth(32)
		linksFrame.create.left.down:SetHeight(32)
		linksFrame.create.left.down:SetPoint("BOTTOMLEFT", 8, 12)
		linksFrame.create.left.down:SetScript("OnClick", function()
			linksFrame.create.left.offset = linksFrame.create.left.offset + 12
			updateLinksCreateLeftButtons(linksFrame.create.left.offset)
		end)
		linksFrame.create.left.func = function(self)
			linksFrame.create.left.ID = self.ID
			linksFrame.create.left.label:SetText("|cFFFFFF00Project:|r " .. potentialLinks[self.ID])
			if linksFrame.create.right.ID ~= 0 then
				linksFrame.create.button:Enable()
			end
		end
		linksFrame.create.left.button = {}
		for i = 1, 12 do
			linksFrame.create.left.button[i] = CreateFrame("Button", "ExcavatinatorLinksCreateLeftButton"..i, linksFrame.create.left)
			linksFrame.create.left.button[i]:SetNormalFontObject('GameFontHighlightSmallLeft')
			linksFrame.create.left.button[i]:SetDisabledFontObject('GameFontHighlightSmallLeft')
			linksFrame.create.left.button[i]:SetWidth(225)
			linksFrame.create.left.button[i]:SetHeight(10)
			linksFrame.create.left.button[i]:SetPushedTextOffset(0, 0)
			linksFrame.create.left.button[i]:SetPoint("TOPLEFT", 44, - (12 * i) - 10)
			linksFrame.create.left.button[i]:SetText("Left button " .. i)
			linksFrame.create.left.button[i].ID = 0
			linksFrame.create.left.button[i]:SetScript("OnClick", linksFrame.create.left.func)
		end
		linksFrame.create.left.label = makeLabel(linksFrame.create.left, 10, "|cFFFFFF00Project:|r <none selected>")
		linksFrame.create.left.label:SetPoint("BOTTOMLEFT", 44, -8)
		linksFrame.create.left.label:SetWidth(200)
		linksFrame.create.left.label:SetHeight(20)
		linksFrame.create.left.label:SetJustifyH("LEFT")
		linksFrame.create.left.label:SetJustifyV("TOP")
		linksFrame.create.left.offset = 0
		linksFrame.create.left.ID = 0
		linksFrame.create.right = CreateFrame("Frame", "ExcavatinatorLinksCreateRightFrame", linksFrame.create)
		linksFrame.create.right:SetWidth(250)
		linksFrame.create.right:SetHeight(180)
		linksFrame.create.right:SetPoint("TOPLEFT", linksFrame.create.left, "TOPRIGHT")
		linksFrame.create.right.up = CreateFrame("Button", "ExcavatinatorLinksCreateUpButton", linksFrame.create.right)
		linksFrame.create.right.up:SetNormalTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Up")
		linksFrame.create.right.up:SetPushedTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Down")
		linksFrame.create.right.up:SetDisabledTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Disabled")
		linksFrame.create.right.up:SetHighlightTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Highlight")
		linksFrame.create.right.up:SetWidth(32)
		linksFrame.create.right.up:SetHeight(32)
		linksFrame.create.right.up:SetPoint("TOPLEFT", 8, -20)
		linksFrame.create.right.up:SetScript("OnClick", function()
			linksFrame.create.right.offset = max(linksFrame.create.right.offset - 12, 0)
			updateLinksCreateRightButtons(linksFrame.items, linksFrame.create.right.offset)
		end)
		linksFrame.create.right.down = CreateFrame("Button", "ExcavatinatorLinksCreateDownButton", linksFrame.create.right)
		linksFrame.create.right.down:SetNormalTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Up")
		linksFrame.create.right.down:SetPushedTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Down")
		linksFrame.create.right.down:SetDisabledTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Disabled")
		linksFrame.create.right.down:SetHighlightTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Highlight")
		linksFrame.create.right.down:SetWidth(32)
		linksFrame.create.right.down:SetHeight(32)
		linksFrame.create.right.down:SetPoint("BOTTOMLEFT", 8, 12)
		linksFrame.create.right.down:SetScript("OnClick", function()
			linksFrame.create.right.offset = linksFrame.create.right.offset + 12
			updateLinksCreateRightButtons(linksFrame.items, linksFrame.create.right.offset)
		end)
		linksFrame.create.right.func = function(self)
			linksFrame.create.right.ID = self.ID
			linksFrame.create.right.label:SetText("|cFFFFFF00Item:|r " .. linksFrame.items[self.ID])
			if linksFrame.create.left.ID ~= 0 then
				linksFrame.create.button:Enable()
			end
		end
		linksFrame.create.right.button = {}
		for i = 1, 12 do
			linksFrame.create.right.button[i] = CreateFrame("Button", "ExcavatinatorLinksCreateLeftButton"..i, linksFrame.create.right)
			linksFrame.create.right.button[i]:SetNormalFontObject('GameFontHighlightSmallLeft')
			linksFrame.create.right.button[i]:SetDisabledFontObject('GameFontHighlightSmallLeft')
			linksFrame.create.right.button[i]:SetWidth(225)
			linksFrame.create.right.button[i]:SetHeight(10)
			linksFrame.create.right.button[i]:SetPushedTextOffset(0, 0)
			linksFrame.create.right.button[i]:SetPoint("TOPLEFT", 44, - (12 * i) - 10)
			linksFrame.create.right.button[i]:SetText("Right button " .. i)
			linksFrame.create.right.button[i].ID = 0
			linksFrame.create.right.button[i]:SetScript("OnClick", linksFrame.create.right.func)
		end
		linksFrame.create.right.label = makeLabel(linksFrame.create.right, 10, "|cFFFFFF00Item:|r <none selected>")
		linksFrame.create.right.label:SetPoint("BOTTOMLEFT", 44, -8)
		linksFrame.create.right.label:SetWidth(200)
		linksFrame.create.right.label:SetHeight(20)
		linksFrame.create.right.label:SetJustifyH("LEFT")
		linksFrame.create.right.label:SetJustifyV("TOP")
		linksFrame.create.right.offset = 0
		linksFrame.create.right.ID = 0
		linksFrame.create.button = CreateFrame("Button", "ExcavatinatorCreateLinkButton", linksFrame.create, "OptionsButtonTemplate")
		linksFrame.create.button:SetPoint("BOTTOM", 0, 8)
		linksFrame.create.button:SetText("Link!")
		linksFrame.create.button:Disable()
		linksFrame.create.button:SetScript("OnClick", function(self)
			self.projectID = linksFrame.create.left.ID
			self.itemID = linksFrame.create.right.ID
			linksFrame.confirm.ok:SetScript("OnClick", linksFrame.confirm.okcreatefunc)
			linksFrame.confirm.self = self
			linksFrame.confirm.label:SetText("Create a link between \"" .. potentialLinks[self.projectID] .. "\" and \"" .. linksFrame.items[self.itemID] .. "\"?")
			linksFrame.confirm:Show()
		end)
	end
	linksFrame:Show()
	linksFrame.modify:Hide()
	linksFrame.modify.offset = 0
	linksFrame.create:Show()
	linksFrame.create.left.offset = 0
	linksFrame.create.right.offset = 0
	linksFrame.links = orderedLinksList()
	linksFrame.items = orderedRemainingArtifactsList()
	updateLinksDeleteButtons(linksFrame.links, 0)
	updateLinksCreateLeftButtons(0)
	updateLinksCreateRightButtons(linksFrame.items, 0)
end

local function openItemsFrame()
	if travelFrame then travelFrame:Hide() end
	if not itemsFrame then
		itemsFrame = CreateFrame("Frame", "ExcavatinatorItemsFrame", raceFrame)
		itemsFrame:SetBackdrop( {
			bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
			edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
			tile = true, tileSize = 16, edgeSize = 16,
			insets = { left = 5, right = 5, top = 5, bottom = 5 } } )
		itemsFrame:SetPoint("TOPRIGHT", raceFrame, "TOPLEFT")
		itemsFrame:SetWidth(48)
		itemsFrame:SetHeight(48)
	--	itemsFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	--	itemsFrame:SetScript("OnEvent", function(self, event, unit) if unit == "player" then self:Hide() end end)
		itemsFrame.buttons = {}
		for i=1, #onUseItems do
			local texture = GetItemIcon(onUseItems[i][1])
			itemsFrame.buttons[i] = makeSecureButton("ExcavatinatorItemButton"..i, itemsFrame, texture, true, onUseItems[i][1], onUseItems[i][2], onUseItems[i][3])
			itemsFrame.buttons[i].itemID = onUseItems[i][1]
			itemsFrame.buttons[i].consumed = onUseItems[i][4]
			local _, link = GetItemInfo(onUseItems[i][1])
			itemsFrame.buttons[i]:SetScript("OnEnter", function()
				GameTooltip:SetOwner(itemsFrame.buttons[i], "ANCHOR_CURSOR")
				GameTooltip:ClearLines()
				GameTooltip:SetHyperlink(link)
			end)
			itemsFrame.buttons[i]:SetScript("OnLeave", function() GameTooltip:Hide() end)
		end
	elseif itemsFrame:IsShown() then
		itemsFrame:Hide()
	else
		itemsFrame:Show()
	end
	local n = 0
	for i=1, #itemsFrame.buttons do
		local count = GetItemCount(itemsFrame.buttons[i].itemID)
		if count > 0 then
			itemsFrame.buttons[i]:Show()
			itemsFrame.buttons[i]:SetPoint("TOPRIGHT", -8 - (36 * math.floor(n / TRAVELFRAME_MAXHEIGHT)), -8 -(36 * (n % TRAVELFRAME_MAXHEIGHT)))
			n = n + 1
			local start, duration = GetItemCooldown(itemsFrame.buttons[i].itemID)
			if duration > 0 then
				itemsFrame.buttons[i].cooldown:SetCooldown(start, duration)
			end
			if itemsFrame.buttons[i].consumed then
				itemsFrame.buttons[i].label:Show()
				itemsFrame.buttons[i].label:SetText(count)
			else
				itemsFrame.buttons[i].label:Hide()
			end
		else
			itemsFrame.buttons[i]:Hide()
		end
	end

	itemsFrame:SetWidth(max(36 * math.floor(((n - 1) / TRAVELFRAME_MAXHEIGHT) + 1) + 12, 48))
	itemsFrame:SetHeight(max(36 * min(n, TRAVELFRAME_MAXHEIGHT) + 12, 48))
end

local function openOptionsFrame()
	if achieveFrame then achieveFrame:Hide() end
	if not optionsFrame then
		optionsFrame = CreateFrame("Frame", "ExcavatinatorOptionsFrame", raceFrame)
		optionsFrame:SetBackdrop( {
			bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
			edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
			tile = true, tileSize = 16, edgeSize = 16,
			insets = { left = 5, right = 5, top = 5, bottom = 5 } } )
		optionsFrame:SetWidth(550)
		optionsFrame:SetHeight(100)
		optionsFrame:SetPoint("TOPLEFT", raceFrame, "BOTTOMLEFT")

		optionsFrame.close = CreateFrame("Button", "ExcavatinatorOptionsFrameCloseButton", optionsFrame, "UIPanelCloseButton")
		optionsFrame.close:SetPoint("TOPRIGHT", -8, -8)

		optionsFrame.minimapShow = CreateFrame("CheckButton", "ExcavatinatorOptionsCheck1", optionsFrame, "OptionsCheckButtonTemplate")
		optionsFrame.minimapShow:SetPoint("TOPLEFT", 8, -8)
		ExcavatinatorOptionsCheck1Text:SetText("Show digs on minimap")
		optionsFrame.minimapShow:SetChecked(Excavatinator:GetMinimapDisplay())
		optionsFrame.minimapShow:SetScript("OnClick", function(self)
			Excavatinator:SetMinimapDisplay(self:GetChecked())
		end)
		optionsFrame.minimapTrack = CreateFrame("CheckButton", "ExcavatinatorOptionsCheck2", optionsFrame, "OptionsCheckButtonTemplate")
		optionsFrame.minimapTrack:SetPoint("TOPLEFT", 8, -32)
		ExcavatinatorOptionsCheck2Text:SetText("Track digs")
		optionsFrame.minimapTrack:SetChecked(Excavatinator:GetMinimapTracking())
		optionsFrame.minimapTrack:SetScript("OnClick", function(self)
			Excavatinator:SetMinimapTracking(self:GetChecked())
		end)

		optionsFrame.links = CreateFrame("Button", "ExcavatinatorLinksButton", optionsFrame)
		optionsFrame.links.dtexture = optionsFrame.links:CreateTexture()
		optionsFrame.links.dtexture:SetTexture("Interface/Icons/INV_Belt_18")
		optionsFrame.links.dtexture:SetAllPoints()
		optionsFrame.links.dtexture:SetDesaturated(true)
		optionsFrame.links.ctexture = optionsFrame.links:CreateTexture()
		optionsFrame.links.ctexture:SetTexture("Interface/Buttons/UI-Quickslot-Depress")
		optionsFrame.links.ctexture:SetAllPoints()
		optionsFrame.links.ctexture:SetBlendMode("ADD")
		optionsFrame.links.ctexture:SetDrawLayer("BORDER")
		optionsFrame.links.htexture = optionsFrame.links:CreateTexture()
		optionsFrame.links.htexture:SetTexture("Interface/Buttons/ButtonHilight-Square")
		optionsFrame.links.htexture:SetAllPoints()
		optionsFrame.links.htexture:SetBlendMode("ADD")
		optionsFrame.links.htexture:SetDrawLayer("HIGHLIGHT")
		optionsFrame.links.ntexture = optionsFrame.links:CreateTexture()
		optionsFrame.links.ntexture:SetTexture("Interface/Icons/INV_Belt_18")
		optionsFrame.links.ntexture:SetDrawLayer("BACKGROUND")
		optionsFrame.links.ntexture:SetAllPoints()
		optionsFrame.links:SetPushedTexture(optionsFrame.links.ctexture)
		optionsFrame.links:SetDisabledTexture(optionsFrame.links.dtexture)
		optionsFrame.links:SetWidth(32)
		optionsFrame.links:SetHeight(32)
		optionsFrame.links:SetPoint("BOTTOMLEFT", 8, 8)
		optionsFrame.links:SetScript("OnClick", openLinksFrame)
		optionsFrame.links:SetScript("OnEnter", function()
			GameTooltip:SetOwner(optionsFrame.links, "ANCHOR_CURSOR")
			GameTooltip:ClearLines()
			GameTooltip:AddLine("Links frame")
			GameTooltip:AddLine("Click to open the links frame.", 1, 1, 1)
			GameTooltip:AddLine("Here you can manually link an archaeology project", 1, 1, 1)
			GameTooltip:AddLine("to the item it creates, if Excavatinator could", 1, 1, 1)
			GameTooltip:AddLine("not do it automatically. You can also delete any", 1, 1, 1)
			GameTooltip:AddLine("existing links, should you choose to.", 1, 1, 1)
			GameTooltip:Show()
		end)
		optionsFrame.links:SetScript("OnLeave", function()
			GameTooltip:Hide()
		end)
	end
	optionsFrame:Show()
end

excavatinator = function()
	if not InCombatLockdown() then
		if not raceFrame then
			-- Load item IDs and turn them into item names
			local newIDs = {} -- Localised item names and IDs
			for i=1, artifactIDs.races do
				artifacts[i] = {
					common = {},
					rare = {},
					special = { n = 0 },
					progress = {
						totalArtifacts = 0,
						completedArtifacts = 0,
						totalSolved = 0,
						currentName = "",
						currentProgress = 0,
						currentMax = 0,
						numKeystones = 0,
						numSlots = 0,
					},
				}
				newIDs[i] = { common = {}, rare = {}, special = {} }
				for k,v in pairs(artifactIDs[i].common) do
					local name = GetItemInfo(v)
					if not name then
						name = k
					end
					artifacts[i].common[name] = false
					newIDs[i].common[name] = v
				end
				for k,v in pairs(artifactIDs[i].rare) do
					local name = GetItemInfo(v)
					if not name then
						name = k
					end
					artifacts[i].rare[name] = false
					newIDs[i].rare[name] = v
				end
				for k,v in pairs(artifactIDs[i].special) do
					if k ~= "n" then
						artifacts[i].special.n = artifacts[i].special.n + 1
						local name = GetItemInfo(v)
						if not name then
							name = k
						end
						artifacts[i].special[name] = -1
						newIDs[i].special[name] = v
					end
				end
			end
			artifacts.races = artifactIDs.races
			newIDs.races = artifactIDs.races
			artifactIDs = newIDs

			-- Load patched item IDs and do the same for them.
			newIDs = {}
			for a,b in pairs(artifactsPatchedIDs) do
				newIDs[a] = {}
				artifactsPatched[a] = {}
				for i=1, b.races do
					artifactsPatched[a][i] = {
						common = {},
						rare = {},
						special = { n = 0 },
					}
					newIDs[a][i] = { common = {}, rare = {}, special = {} }
					for k,v in pairs(b[i].common) do
						local name = GetItemInfo(v)
						if not name then
							name = k
						end
						artifactsPatched[a][i].common[name] = false
						newIDs[a][i].common[name] = v
					end
					for k,v in pairs(b[i].rare) do
						local name = GetItemInfo(v)
						if not name then
							name = k
						end
						artifactsPatched[a][i].rare[name] = false
						newIDs[a][i].rare[name] = v
					end
					for k,v in pairs(b[i].special) do
						if k ~= "n" then
							artifactsPatched[a][i].special.n = artifactsPatched[a][i].special.n + 1
							local name = GetItemInfo(v)
							if not name then
								name = k
							end
							artifactsPatched[a][i].special[name] = -1
							newIDs[a][i].special[name] = v
						end
					end
				end
				artifactsPatched[a].patch = b.patch
				artifactsPatched[a].races = b.races
				newIDs[a].patch = b.patch
				newIDs[a].races = b.races
			end
			artifactsPatchedIDs = newIDs

			-- Add any patched-in artifacts to the list:
			for k,v in pairs(artifactsPatched) do
				if patchVersion >= k then
					for i=1, v.races do
						for kk, vv in pairs(v[i].common) do
							artifacts[i].common[kk] = vv
						end
						for kk, vv in pairs(v[i].rare) do
							artifacts[i].rare[kk] = vv
						end
						if v[i].special then
							for kk, vv in pairs(v[i].special) do
								if kk ~= "n" then
									artifacts[i].special[kk] = vv
								end
							end
						end
					end
					if v.races > artifacts.races then artifacts.races = v.races end
				end
			end

			-- Load patched item IDs and do the same for them.
			for a,b in pairs(artifactsPatchedIDs) do
				for i=1, b.races do
					if b.races > artifactIDs.races then
						for j=artifactIDs.races+1, b.races do
							artifactIDs[j] = {
								common = {},
								rare = {},
								special = { n = 0 },
							}
						end
						artifactIDs.races = b.races
					end
					for k,v in pairs(b[i].common) do
						artifactIDs[i].common[k] = v
					end
					for k,v in pairs(b[i].rare) do
						artifactIDs[i].rare[k] = v
					end
					for k,v in pairs(b[i].special) do
						artifactIDs[i].special.n = artifactIDs[i].special.n + 1
						artifactIDs[i].special[k] = v
					end
				end
			end

			historyFrame:RegisterEvent("ARTIFACT_HISTORY_READY")

			raceFrame = CreateFrame("Frame", "ExcavatinatorRaceFrame", UIParent)
			raceFrame.isCollapsed = false
			raceFrame:EnableMouse(true)
			raceFrame:SetMovable(true)
			raceFrame:SetScript("OnDragStart", raceFrame.StartMoving)
			raceFrame:SetScript("OnDragStop", raceFrame.StopMovingOrSizing)
			raceFrame:RegisterForDrag("LeftButton")
			raceFrame:SetBackdrop( {
				bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
				edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
				tile = true, tileSize = 16, edgeSize = 16,
				insets = { left = 5, right = 5, top = 5, bottom = 5 } } )
			raceFrame:SetWidth(550)
			raceFrame:SetHeight(255)
			raceFrame:SetPoint("CENTER")

			raceFrame.button = {}
			raceFrame.progressBar = {}
			raceFrame.currentBar = {}
			raceFrame.keystones = {}

			raceFrame.header1 = makeLabel(raceFrame, 14, "|c00FFFF00Race|r")
			raceFrame.header1:SetPoint("TOPLEFT", 12, -12)
			raceFrame.header2 = makeLabel(raceFrame, 14, "|c00FFFF00Progress|r")
			raceFrame.header2:SetPoint("TOPLEFT", 120, -12)
			raceFrame.header3 = makeLabel(raceFrame, 14, "|c00FFFF00Artifact|r")
			raceFrame.header3:SetPoint("TOPLEFT", 300, -12)
			raceFrame.header4 = makeLabel(raceFrame, 14, "|c00FFFF00Keystones|r")
			raceFrame.header4:SetPoint("TOPRIGHT", -32, -12)

			for i=1,GetNumArchaeologyRaces()-1 do
				-- Count the total number of artifacts and the number of completed artifacts for each race

				local total = 0
				for k,v in pairs(artifacts[i].common) do
					total = total + 1
				end
				for k,v in pairs(artifacts[i].rare) do
					total = total + 1
				end
				artifacts[i].progress.totalArtifacts = total

				raceFrame.button[i] = CreateFrame("Button", "ExcavatinatorButton"..i, raceFrame)
				raceFrame.button[i]:SetNormalFontObject('GameFontHighlightLeft')
				raceFrame.button[i]:SetText(GetArchaeologyRaceInfo(i))
				raceFrame.button[i]:SetWidth(500)
				raceFrame.button[i]:SetHeight(20)
				raceFrame.button[i]:SetPoint("TOPLEFT", 12, -12-20*i)
				raceFrame.button[i]:SetScript("OnClick", function(self, button)
					if button == "LeftButton" then
						populateScrollFrame(i)
					elseif button == "RightButton" then
						solveArtifact(i)
					end
				end)
				raceFrame.button[i]:SetScript("OnEnter", function()
					GameTooltip:SetOwner(raceFrame.button[i], "ANCHOR_CURSOR")
					GameTooltip:ClearLines()
					local race = GetArchaeologyRaceInfo(i)
					GameTooltip:AddLine(race)
					GameTooltip:AddLine(format("%d / %d artifacts completed.", artifacts[i].progress.completedArtifacts, artifacts[i].progress.totalArtifacts), 1, 1, 1)
					GameTooltip:AddLine(format("%d solves total.", artifacts[i].progress.totalSolved), 1, 1, 1)
					GameTooltip:AddLine(" ")
					GameTooltip:AddDoubleLine("Project" .. ":", artifacts[i].progress.currentName, 1, 1, 0, 1, 1, 1)
					local req = max(artifacts[i].progress.currentMax - artifacts[i].progress.currentProgress - 12*min(artifacts[i].progress.numKeystones, artifacts[i].progress.numSlots), 0)
					GameTooltip:AddDoubleLine(
						format("%s / %d fragments",
						(min(artifacts[i].progress.numKeystones, artifacts[i].progress.numSlots) >= 1 and
						artifacts[i].progress.currentProgress .. "+" .. 12*min(artifacts[i].progress.numKeystones, artifacts[i].progress.numSlots)
						or artifacts[i].progress.currentProgress), artifacts[i].progress.currentMax),
						req == 0 and "Ready to solve." or format("Need %d more.", req),
						1, 1, 1, req == 0 and 0 or 1, req == 0 and 1 or 0, 0)
					GameTooltip:AddLine(" ")
					GameTooltip:AddLine("Click to open artifact list.", 0, 1, 0)
					if req == 0 then GameTooltip:AddLine("Right click to solve artifact.", 0, 1, 0) end
					GameTooltip:Show()
				end)
				raceFrame.button[i]:SetScript("OnLeave", function() GameTooltip:Hide() end)
				raceFrame.button[i]:RegisterForClicks("LeftButtonUp", "RightButtonUp")

				raceFrame.progressBar[i] = CreateFrame("Frame", "ExcavatinatorProgressBar"..i, raceFrame)
				raceFrame.progressBar[i]:SetWidth(80)
				raceFrame.progressBar[i]:SetHeight(10)
				raceFrame.progressBar[i]:SetPoint("TOPLEFT", 120, -17-20*i)
				setTexture(raceFrame.progressBar[i], 0.3, 0.3, 0.3)
				raceFrame.progressBar[i].bar = CreateFrame("Frame", "ExcavatinatorProgressBarFill"..i, raceFrame.progressBar[i])
				raceFrame.progressBar[i].bar:SetWidth(40)
				raceFrame.progressBar[i].bar:SetHeight(10)
				raceFrame.progressBar[i].bar:SetPoint("TOPLEFT")
				setTexture(raceFrame.progressBar[i].bar, 0, 1, 0)
				raceFrame.progressBar[i].label = makeLabel(raceFrame.progressBar[i], 12, "25/50")
				raceFrame.progressBar[i].label:SetPoint("LEFT", raceFrame.progressBar[i], "RIGHT", 8, 0)

				raceFrame.currentBar[i] = CreateFrame("Frame", "ExcavatinatorCurrentBar"..i, raceFrame)
				raceFrame.currentBar[i]:SetWidth(80)
				raceFrame.currentBar[i]:SetHeight(10)
				raceFrame.currentBar[i]:SetPoint("TOPLEFT", 300, -17-20*i)
				setTexture(raceFrame.currentBar[i], 0.3, 0.3, 0.3)
				raceFrame.currentBar[i].fragmentBar = CreateFrame("Frame", "ExcavatinatorCurrentFragmentBar"..i, raceFrame.currentBar[i])
				raceFrame.currentBar[i].fragmentBar:SetWidth(30)
				raceFrame.currentBar[i].fragmentBar:SetHeight(10)
				raceFrame.currentBar[i].fragmentBar:SetPoint("TOPLEFT")
				setTexture(raceFrame.currentBar[i].fragmentBar, 0, 1, 0)
				raceFrame.currentBar[i].keystoneBar = CreateFrame("Frame", "ExcavatinatorCurrentKeystoneBar"..i, raceFrame.currentBar[i])
				raceFrame.currentBar[i].keystoneBar:SetWidth(30)
				raceFrame.currentBar[i].keystoneBar:SetHeight(10)
				raceFrame.currentBar[i].keystoneBar:SetPoint("TOPLEFT", raceFrame.currentBar[i].fragmentBar, "TOPRIGHT")
				setTexture(raceFrame.currentBar[i].keystoneBar, 1, 1, 0)
				raceFrame.currentBar[i].label = makeLabel(raceFrame.currentBar[i], 12, "15+15/40")
				raceFrame.currentBar[i].label:SetPoint("LEFT", raceFrame.currentBar[i], "RIGHT", 8, 0)

				raceFrame.keystones[i] = makeLabel(raceFrame, 12, "3")
				raceFrame.keystones[i]:SetPoint("TOPLEFT", 500, -16-20*i)

			end
			raceFrame.close = CreateFrame("Button", "ExcavatinatorCloseRaceFrameButton", raceFrame, "UIPanelCloseButton")
			raceFrame.close:SetPoint("TOPRIGHT", -8, -8)

			raceFrame.expand = CreateFrame("Button", "ExcavatinatorExpandButton", raceFrame)
			raceFrame.expand:SetNormalTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Up")
			raceFrame.expand:SetPushedTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Down")
			raceFrame.expand:SetDisabledTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Disabled")
			raceFrame.expand:SetHighlightTexture("Interface/Buttons/UI-ScrollBar-ScrollDownButton-Highlight")
			raceFrame.expand:SetWidth(32)
			raceFrame.expand:SetHeight(32)
			raceFrame.expand:SetPoint("BOTTOMRIGHT", -4, 4)
			raceFrame.expand:SetScript("OnClick", expand)
			raceFrame.expand:Hide()
			raceFrame.collapse = CreateFrame("Button", "ExcavatinatorCollapseButton", raceFrame)
			raceFrame.collapse:SetNormalTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Up")
			raceFrame.collapse:SetPushedTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Down")
			raceFrame.collapse:SetDisabledTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Disabled")
			raceFrame.collapse:SetHighlightTexture("Interface/Buttons/UI-ScrollBar-ScrollUpButton-Highlight")
			raceFrame.collapse:SetWidth(32)
			raceFrame.collapse:SetHeight(32)
			raceFrame.collapse:SetPoint("BOTTOMRIGHT", -4, 4)
			raceFrame.collapse:SetScript("OnClick", collapse)

			raceFrame.survey = makeSecureButton("ExcavatinatorSurveyButton", raceFrame, "INTERFACE/ICONS/inv_misc_shovel_01", false, nil, 80451, 3)
			raceFrame.survey:RegisterEvent("LEARNED_SPELL_IN_TAB")
			raceFrame.survey:SetScript("OnEvent", surveyButton_onEvent)
			raceFrame.survey:SetPoint("BOTTOM", -20, 8)
			raceFrame.survey:SetScript("OnEnter", function()
				GameTooltip:SetOwner(raceFrame.survey, "ANCHOR_CURSOR")
				GameTooltip:ClearLines()
				GameTooltip:SetHyperlink(GetSpellLink(80451))
				GameTooltip:Show()
			end)
			raceFrame.survey:SetScript("OnLeave", function()
				GameTooltip:Hide()
			end)

			raceFrame.achievement = CreateFrame("Button", "ExcavatinatorAchievementButton", raceFrame)
			raceFrame.achievement.ntexture = raceFrame.achievement:CreateTexture()
			raceFrame.achievement.ntexture:SetTexture("Interface/Buttons/UI-MicroButton-Achievement-Up")
			raceFrame.achievement.ntexture:SetAllPoints()
			raceFrame.achievement.ntexture:SetTexCoord(0.07, 0.93, 0.38, 0.98)
			raceFrame.achievement:SetNormalTexture(raceFrame.achievement.ntexture)
			raceFrame.achievement.htexture = raceFrame.achievement:CreateTexture()
			raceFrame.achievement.htexture:SetTexture("Interface/Buttons/ButtonHilight-Square")
			raceFrame.achievement.htexture:SetAllPoints()
			raceFrame.achievement.htexture:SetBlendMode("ADD")
			raceFrame.achievement.htexture:SetDrawLayer("HIGHLIGHT")
			raceFrame.achievement.ctexture = raceFrame.achievement:CreateTexture()
			raceFrame.achievement.ctexture:SetTexture("Interface/Buttons/UI-MicroButton-Achievement-Down")
			raceFrame.achievement.ctexture:SetAllPoints()
			raceFrame.achievement.ctexture:SetTexCoord(0.07, 0.93, 0.38, 0.98)
			raceFrame.achievement:SetPushedTexture(raceFrame.achievement.ctexture)
			raceFrame.achievement:SetWidth(32)
			raceFrame.achievement:SetHeight(32)
			raceFrame.achievement:SetPoint("BOTTOM", 20, 8)
			raceFrame.achievement:SetScript("OnClick", updateAchievements)
			raceFrame.achievement:SetScript("OnEnter", function()
				GameTooltip:SetOwner(raceFrame.achievement, "ANCHOR_CURSOR")
				GameTooltip:ClearLines()
				GameTooltip:AddLine("Achievement tracker")
				GameTooltip:AddLine("The achievement tracker provides you with", 1, 1, 1)
				GameTooltip:AddLine("information about the archaeology achievements,", 1, 1, 1)
				GameTooltip:AddLine("and your progress for them.", 1, 1, 1)
				GameTooltip:Show()
			end)
			raceFrame.achievement:SetScript("OnLeave", function()
				GameTooltip:Hide()
			end)

			raceFrame.travel = CreateFrame("Button", "ExcavatinatorTravelButton", raceFrame)
			raceFrame.travel.dtexture = raceFrame.travel:CreateTexture()
			raceFrame.travel.dtexture:SetTexture("Interface/Icons/Ability_Mount_FossilizedRaptor")
			raceFrame.travel.dtexture:SetAllPoints()
			raceFrame.travel.dtexture:SetDesaturated(true)
			raceFrame.travel.ctexture = raceFrame.travel:CreateTexture()
			raceFrame.travel.ctexture:SetTexture("Interface/Buttons/UI-Quickslot-Depress")
			raceFrame.travel.ctexture:SetAllPoints()
			raceFrame.travel.ctexture:SetBlendMode("ADD")
			raceFrame.travel.ctexture:SetDrawLayer("BORDER")
			raceFrame.travel.htexture = raceFrame.travel:CreateTexture()
			raceFrame.travel.htexture:SetTexture("Interface/Buttons/ButtonHilight-Square")
			raceFrame.travel.htexture:SetAllPoints()
			raceFrame.travel.htexture:SetBlendMode("ADD")
			raceFrame.travel.htexture:SetDrawLayer("HIGHLIGHT")
			raceFrame.travel.ntexture = raceFrame.travel:CreateTexture()
			raceFrame.travel.ntexture:SetTexture("Interface/Icons/Ability_Mount_FossilizedRaptor")
			raceFrame.travel.ntexture:SetDrawLayer("BACKGROUND")
			raceFrame.travel.ntexture:SetAllPoints()
			raceFrame.travel:SetPushedTexture(raceFrame.travel.ctexture)
			raceFrame.travel:SetDisabledTexture(raceFrame.travel.dtexture)
			raceFrame.travel:SetWidth(32)
			raceFrame.travel:SetHeight(32)
			raceFrame.travel:SetPoint("BOTTOMLEFT", 8, 8)
			raceFrame.travel:SetScript("OnClick", openTravelFrame)
			raceFrame.travel:SetScript("OnEnter", function()
				GameTooltip:SetOwner(raceFrame.travel, "ANCHOR_CURSOR")
				GameTooltip:ClearLines()
				GameTooltip:AddLine("Travel frame")
				GameTooltip:AddLine("Click to open a list of any instant-travel item", 1, 1, 1)
				GameTooltip:AddLine("or ability you have.", 1, 1, 1)
				GameTooltip:Show()
			end)
			raceFrame.travel:SetScript("OnLeave", function()
				GameTooltip:Hide()
			end)

			raceFrame.items = CreateFrame("Button", "ExcavatinatorItemsButton", raceFrame)
			raceFrame.items.dtexture = raceFrame.items:CreateTexture()
			raceFrame.items.dtexture:SetTexture("Interface/Icons/inv_misc_archaeology_vrykuldrinkinghorn")
			raceFrame.items.dtexture:SetAllPoints()
			raceFrame.items.dtexture:SetDesaturated(true)
			raceFrame.items.ctexture = raceFrame.items:CreateTexture()
			raceFrame.items.ctexture:SetTexture("Interface/Buttons/UI-Quickslot-Depress")
			raceFrame.items.ctexture:SetAllPoints()
			raceFrame.items.ctexture:SetBlendMode("ADD")
			raceFrame.items.ctexture:SetDrawLayer("BORDER")
			raceFrame.items.htexture = raceFrame.items:CreateTexture()
			raceFrame.items.htexture:SetTexture("Interface/Buttons/ButtonHilight-Square")
			raceFrame.items.htexture:SetAllPoints()
			raceFrame.items.htexture:SetBlendMode("ADD")
			raceFrame.items.htexture:SetDrawLayer("HIGHLIGHT")
			raceFrame.items.ntexture = raceFrame.items:CreateTexture()
			raceFrame.items.ntexture:SetTexture("Interface/Icons/inv_misc_archaeology_vrykuldrinkinghorn")
			raceFrame.items.ntexture:SetDrawLayer("BACKGROUND")
			raceFrame.items.ntexture:SetAllPoints()
			raceFrame.items:SetPushedTexture(raceFrame.items.ctexture)
			raceFrame.items:SetDisabledTexture(raceFrame.items.dtexture)
			raceFrame.items:SetWidth(32)
			raceFrame.items:SetHeight(32)
			raceFrame.items:SetPoint("BOTTOMLEFT", 48, 8)
			raceFrame.items:SetScript("OnClick", openItemsFrame)
			raceFrame.items:SetScript("OnEnter", function()
				GameTooltip:SetOwner(raceFrame.items, "ANCHOR_CURSOR")
				GameTooltip:ClearLines()
				GameTooltip:AddLine("Items frame")
				GameTooltip:AddLine("Click to open a list of any archaeology", 1, 1, 1)
				GameTooltip:AddLine("on-use items you have.", 1, 1, 1)
				GameTooltip:Show()
			end)
			raceFrame.items:SetScript("OnLeave", function()
				GameTooltip:Hide()
			end)

			raceFrame.options = CreateFrame("Button", "ExcavatinatorOptionsButton", raceFrame)
			raceFrame.options.dtexture = raceFrame.options:CreateTexture()
			raceFrame.options.dtexture:SetTexture("Interface/Icons/Trade_Engineering")
			raceFrame.options.dtexture:SetAllPoints()
			raceFrame.options.dtexture:SetDesaturated(true)
			raceFrame.options.ctexture = raceFrame.options:CreateTexture()
			raceFrame.options.ctexture:SetTexture("Interface/Buttons/UI-Quickslot-Depress")
			raceFrame.options.ctexture:SetAllPoints()
			raceFrame.options.ctexture:SetBlendMode("ADD")
			raceFrame.options.ctexture:SetDrawLayer("BORDER")
			raceFrame.options.htexture = raceFrame.options:CreateTexture()
			raceFrame.options.htexture:SetTexture("Interface/Buttons/ButtonHilight-Square")
			raceFrame.options.htexture:SetAllPoints()
			raceFrame.options.htexture:SetBlendMode("ADD")
			raceFrame.options.htexture:SetDrawLayer("HIGHLIGHT")
			raceFrame.options.ntexture = raceFrame.options:CreateTexture()
			raceFrame.options.ntexture:SetTexture("Interface/Icons/Trade_Engineering")
			raceFrame.options.ntexture:SetDrawLayer("BACKGROUND")
			raceFrame.options.ntexture:SetAllPoints()
			raceFrame.options:SetPushedTexture(raceFrame.options.ctexture)
			raceFrame.options:SetDisabledTexture(raceFrame.options.dtexture)
			raceFrame.options:SetWidth(32)
			raceFrame.options:SetHeight(32)
			raceFrame.options:SetPoint("BOTTOMRIGHT", -40, 8)
			raceFrame.options:SetScript("OnClick", openOptionsFrame)
			raceFrame.options:SetScript("OnEnter", function()
				GameTooltip:SetOwner(raceFrame.options, "ANCHOR_CURSOR")
				GameTooltip:ClearLines()
				GameTooltip:AddLine("Options frame")
				GameTooltip:AddLine("Click to open the options frame.", 1, 1, 1)
				GameTooltip:Show()
			end)
			raceFrame.options:SetScript("OnLeave", function()
				GameTooltip:Hide()
			end)

			achieveFrame = CreateFrame("Frame", "ExcavatinatorAchieveFrame", raceFrame)
			achieveFrame:SetBackdrop( {
				bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
				edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
				tile = true, tileSize = 16, edgeSize = 16,
				insets = { left = 5, right = 5, top = 5, bottom = 5 } } )
			achieveFrame:SetWidth(550)
			achieveFrame:SetHeight(220)
			achieveFrame:SetPoint("TOPLEFT", raceFrame, "BOTTOMLEFT")

			achieveFrame.button = {}
			achieveFrame.progress = {}

			for i=1, 10 do
				achieveFrame.button[i] = CreateFrame("Button", "ExcavatinatorAchieveButton" .. i, achieveFrame)
				achieveFrame.button[i]:SetNormalFontObject('GameFontHighlightLeft')
				achieveFrame.button[i]:SetText("Achievement " .. i)
				achieveFrame.button[i]:SetWidth(500)
				achieveFrame.button[i]:SetHeight(20)
				achieveFrame.button[i]:SetPoint("TOPLEFT", 12, 8-20*i)
				achieveFrame.button[i]:SetScript("OnEnter", function()
					GameTooltip:SetOwner(achieveFrame.button[i], "ANCHOR_CURSOR")
					GameTooltip:ClearLines()
					GameTooltip:SetHyperlink(GetAchievementLink(achievementIDs[i]))
				end)
				achieveFrame.button[i]:SetScript("OnLeave", function() GameTooltip:Hide() end)
				achieveFrame.button[i]:SetScript("OnClick", function() expandAchievement(i) end)

				achieveFrame.progress[i] = CreateFrame("Frame", "ExcavatinatorAchieveProgress" .. i, achieveFrame)
				achieveFrame.progress[i]:SetWidth(80)
				achieveFrame.progress[i]:SetHeight(10)
				achieveFrame.progress[i]:SetPoint("TOPLEFT", 300, 3-20*i)
				setTexture(achieveFrame.progress[i], 0.3, 0.3, 0.3)
				achieveFrame.progress[i].bar = CreateFrame("Frame", "ExcavatinatorAchieveProgressBarFill"..i, achieveFrame.progress[i])
				achieveFrame.progress[i].bar:SetWidth(40)
				achieveFrame.progress[i].bar:SetHeight(10)
				achieveFrame.progress[i].bar:SetPoint("TOPLEFT")
				setTexture(achieveFrame.progress[i].bar, 0, 1, 0)
				achieveFrame.progress[i].label = makeLabel(achieveFrame.progress[i], 12, "25/50")
				achieveFrame.progress[i].label:SetPoint("LEFT", achieveFrame.progress[i], "RIGHT", 8, 0)
			end
			achieveFrame.header = makeLabel(achieveFrame, 10, "Header")
			achieveFrame.header:SetPoint("TOPLEFT", 32, -32)
			achieveFrame.header:Hide()
			achieveFrame.line = {}
			achieveFrame.rightLine = {}
			for i=1, 9 do
				achieveFrame.line[i] = CreateFrame("Button", "ExcavatinatorAchieveLine" .. i, achieveFrame)
				achieveFrame.line[i]:SetNormalFontObject('GameFontHighlightSmallLeft')
				achieveFrame.line[i]:SetDisabledFontObject('GameFontHighlightSmallLeft')
				achieveFrame.line[i]:SetWidth(350)
				achieveFrame.line[i]:SetHeight(10)
				achieveFrame.line[i]:SetPushedTextOffset(0, 0)
				achieveFrame.line[i]:SetPoint("TOPLEFT", 32, - (12 * i) - 44)
				achieveFrame.line[i]:SetText("Line " .. i)
				achieveFrame.line[i]:Hide()
				achieveFrame.line[i].item = 0
				achieveFrame.line[i]:SetScript("OnEnter", function()
					if achieveFrame.line[i].item ~= 0 then
						GameTooltip:SetOwner(achieveFrame.line[i], "ANCHOR_CURSOR")
						GameTooltip:ClearLines()
						local _, link = GetItemInfo(achieveFrame.line[i].item)
						GameTooltip:SetHyperlink(link)
					end
				end)
				achieveFrame.line[i]:SetScript("OnLeave", function() GameTooltip:Hide() end)
				achieveFrame.rightLine[i] = makeLabel(achieveFrame, 10, "Right line " .. i)
				achieveFrame.rightLine[i]:SetPoint("TOPLEFT", 300, - (12 * i) - 44)
				achieveFrame.rightLine[i]:Hide()
			end
			achieveFrame.close = CreateFrame("Button", "ExcavatinatorCloseAchieveFrameButton", achieveFrame, "UIPanelCloseButton")
			achieveFrame.close:SetPoint("TOPRIGHT", -8, -8)

			achieveFrame.ID = 0
			achieveFrame:Hide()

			ExcavatinatorListFrame:ClearAllPoints()
			ExcavatinatorListFrame:SetPoint("TOPLEFT", raceFrame, "TOPRIGHT")
			ExcavatinatorListFrame.close = CreateFrame("Button", "ExcavatinatorCloseRaceFrameButton", ExcavatinatorListFrame, "UIPanelCloseButton")
			ExcavatinatorListFrame.close:SetPoint("TOPRIGHT", -4, -4)

			raceFrame:SetScript("OnShow", function()
				for i=1, GetNumArchaeologyRaces()-1 do
					local p = artifacts[i].progress
					if p.completedArtifacts == 0 then
						raceFrame.progressBar[i].bar:Hide()
					else
						raceFrame.progressBar[i].bar:Show()
						raceFrame.progressBar[i].bar:SetWidth(min((p.completedArtifacts / (p.totalArtifacts ~= 0 and p.totalArtifacts or 1)), 1) * 80) -- Avoid division by 0
					end
					raceFrame.progressBar[i].label:SetText(p.completedArtifacts .. "/" .. p.totalArtifacts)

					if p.currentProgress == 0 then
						raceFrame.currentBar[i].fragmentBar:Hide()
						raceFrame.currentBar[i].keystoneBar:ClearAllPoints()
						raceFrame.currentBar[i].keystoneBar:SetPoint("TOPLEFT")
					else
						raceFrame.currentBar[i].fragmentBar:Show()
						raceFrame.currentBar[i].fragmentBar:SetWidth(min((p.currentProgress / (p.currentMax ~= 0 and p.currentMax or 1)) * 80, 80)) -- Avoid division by 0
						raceFrame.currentBar[i].keystoneBar:ClearAllPoints()
						raceFrame.currentBar[i].keystoneBar:SetPoint("TOPLEFT", raceFrame.currentBar[i].fragmentBar, "TOPRIGHT")
					end
					if min(p.numKeystones, p.numSlots) == 0 or p.currentProgress >= p.currentMax then
						raceFrame.currentBar[i].keystoneBar:Hide()
					else
						raceFrame.currentBar[i].keystoneBar:Show()
						raceFrame.currentBar[i].keystoneBar:SetWidth((12 * min(p.numKeystones, p.numSlots) / p.currentMax) * 80)
						if raceFrame.currentBar[i].fragmentBar:GetWidth() + raceFrame.currentBar[i].keystoneBar:GetWidth() > 80 then
							raceFrame.currentBar[i].keystoneBar:SetWidth(80-raceFrame.currentBar[i].fragmentBar:GetWidth())
						end
					end

					local fragmentsString = p.currentProgress
					if min(p.numKeystones, p.numSlots) > 0 then
						fragmentsString = fragmentsString .. "+" .. 12 * min(p.numKeystones, p.numSlots)
					end
					fragmentsString = fragmentsString .. "/" .. p.currentMax
					raceFrame.currentBar[i].label:SetText(fragmentsString)

					raceFrame.keystones[i]:SetText(i==3 and " " or p.numKeystones)
				end
			end)
			raceFrame:Hide()
			updateFrame:SetScript("OnEvent", function()
				if raceFrame:IsShown() then
					excavatinator()
					if raceFrame.isCollapsed then
						collapse() -- Update which races are shown
					end
					if achieveFrame:IsShown() then
						local ID = achieveFrame.ID
						updateAchievements()
						if ID ~= 0 then
							expandAchievement(ID)
						end
					end
				end
			end)
		end
		surveyButton_onEvent()
		if historyFrame:IsShown() then
			historyFrame:SetScript("OnEvent", function(self) getProgress(self); raceFrame:Hide(); raceFrame:Show() end)
		else
			historyFrame:SetScript("OnEvent", nil)
		end
		RequestArtifactCompletionHistory()
	end
end


SlashCmdList["EXCAVATINATOR"] = excavatinator

BINDING_HEADER_EXCAVATINATOR_SURVEY_HEADER = "Excavatinator"
_G["BINDING_NAME_SPELL Survey"] = "Survey"
BINDING_NAME_EXCAVATINATOR_OPEN = "Open Excavatinator"

-- Bit of a hack - load all item info to make sure it's available.
for i=1, artifactIDs.races do
	for k,v in pairs(artifactIDs[i]) do
		for kk, vv in pairs(v) do
			GetItemInfo(vv)
		end
	end
end

for k,v in pairs(artifactsPatchedIDs) do
	for i=1, v.races do
		for kk, vv in pairs(v[i]) do
			for kkk, vvv in pairs(vv) do
				GetItemInfo(vvv)
			end
		end
	end
end

local function onLoadOrUnload(self, event, addonName)
	if event == "ADDON_LOADED" and addonName == "Excavatinator" then
		if ExcavatinatorDB then
			linkedProjects = ExcavatinatorDB.linkedProjects
		else
			ExcavatinatorDB = {
				linkedProjects = linkedProjects
			}
		end
		for k,v in pairs(linkedProjects) do
			linkedProjectsByItemName[v] = k
		end
	elseif event == "PLAYER_LOGOUT" then
		ExcavatinatorDB.linkedProjects = linkedProjects
	end
end

local loadFrame = CreateFrame("Frame")
loadFrame:RegisterEvent("ADDON_LOADED")
loadFrame:RegisterEvent("PLAYER_LOGOUT")
loadFrame:SetScript("OnEvent", onLoadOrUnload)

-----------------------------
-- API - support functions --
-----------------------------

local progressCallbackList = {}

getProgressCallback = function()
	for i=1, #progressCallbackList do
		if type(progressCallbackList[i]) == "function" then
			progressCallbackList[i]()
		end
	end
end

local function copyTable(tbl)
	if type(tbl) ~= "table" then
		return tbl
	else
		local r = {}
		for k,v in pairs(tbl) do
			r[k] = copyTable(v)
		end
		return r
	end
end

-----------------------------------
-- API - communication functions --
-----------------------------------

--[[
	These functions can be used to retrieve the information that Excavatinator has gathered.
	I recognise that I am not at all a good UI designer, and that others can do a much
	better job of presenting the list of completed artifacts. The following functions will
	allow said others to make their own addons to list the information without having to
	modify Excavatinator, simply by listing it as a dependency.
]]

Excavatinator = {}

--[[ Excavatinator:RegisterCallback(func)
	Use:
		The function passed will be called whenever Excavatinator has retrieved information
		from the Blizzard API. No arguments will be passed to the function; the function
		should use the other functions provided to get the necessary information.
	Returns:
		The internal callback ID. Use this ID to unregister the callback. ]]
function Excavatinator:RegisterCallback(func)
	progressCallbackList[#progressCallbackList+1] = func
	return #progressCallbackList
end

--[[ Excavatinator:UnregisterCallback(ID)
	Use:
		Use the ID provided by RegisterCallback() to unregister a callback. ]]
function Excavatinator:UnregisterCallback(ID)
	if progressCallbackList[ID] then
		progressCallbackList[ID] = nil
	end
end

--[[ Excavatinator:GetNumArtifactsByRace(raceID)
	Arguments:
		raceID - the ID of the race to get artifact count for. Integer between 1 and GetNumArtifactsByRace()-1.
	Returns:
		Completed artifacts - the number of artifacts completed for a given race
		Total artifacts - the total number of existing artifacts for a given race ]]
function Excavatinator:GetNumArtifactsByRace(raceID)
	return artifacts[raceID].progress.completedArtifacts, artifacts[raceID].progress.totalArtifacts
end

--[[ Excavatinator:GetArtifactList(raceID)
	Arguments:
		raceID - the ID of the race to get the artifact list for. Integer between 1 and GetNumArtifactsByRace()-1.
	Returns:
		Artifact list - names and solved (common and rare: boolean; special: -1 for unknown, 0 for incomplete, 1 for completed) ]]
function Excavatinator:GetArtifactList(raceID)
	return { common = copyTable(artifacts[raceID].common), rare = copyTable(artifacts[raceID].rare), special = copyTable(artifacts[raceID].special or nil) }
end

--[[ Excavatinator:GetArtifactItemIDList(raceID)
	Arguments:
		raceID - the ID of the race to get the artifact list for. Integer between 1 and GetNumArtifactsByRace()-1.
	Returns:
		Artifact list - names and item IDs. ]]
function Excavatinator:GetArtifactItemIDList(raceID)
	return copyTable(artifactIDs)
end